import os
import math
import requests
import json
import datetime
import hashlib
import dateparser
import re
import csv
import time
from pygdpr.models.dpa import DPA
from bs4 import BeautifulSoup
from pygdpr.services.filename_from_path_service import filename_from_path_service
from pygdpr.services.pdf_to_text_service import PDFToTextService
from pygdpr.specifications import pdf_file_extension_specification
from pygdpr.specifications.should_retain_document_specification import ShouldRetainDocumentSpecification
from pygdpr.models.common.pagination import Pagination
from pygdpr.policies.gdpr_policy import GDPRPolicy
import textract
from selenium import webdriver
from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from pygdpr.policies.webdriver_exec_policy import WebdriverExecPolicy
import sys


class Malta(DPA):
    def __init__(self, path=os.curdir):
        country_code='mt'
        super().__init__(country_code, path)

    # Method never called
    def update_pagination(self, pagination=None, page_soup=None, driver=None):
        source = {
            "host": "https://idpc.org.mt",
            "start_path": "https://idpc.org.mt"
        }
        host = source['host']
        start_path = source['start_path']
        if pagination is None:
            page_url = host + start_path
            exec_path = WebdriverExecPolicy().get_system_path()
            options = webdriver.ChromeOptions()
            options.add_argument('headless')
            driver = webdriver.Chrome(options=options, executable_path=exec_path)
            driver.get(page_url)
            pagination = Pagination()
            pagination.add_item(driver)
        else:
            news_pagination = driver.find_element_by_id('news-pagination')
            page_btn = news_pagination.find_element_by_class_name('page-btn')
            old_news_entries = driver.find_elements_by_class_name('news-entry')
            page_btn.click()
            time.sleep(20)
            news_entries = driver.find_elements_by_class_name('news-entry')
            if len(news_entries) > len(old_news_entries):
                pagination = Pagination()
                pagination.add_item(driver)
        return pagination

    def get_source(self, page_url=None, driver=None, to_print=True):
        assert (page_url is not None)
        results_response = None
        try:
            results_response = requests.request('GET', page_url)
            results_response.raise_for_status()
        except requests.exceptions.HTTPError as error:
            if to_print:
                print(error)
            pass
        return results_response

    def get_docs(self, existing_docs=[], overwrite=False, to_print=True):
        added_docs = []
        # call all the get_docs_X() functions
        added_docs += self.get_docs_Guidelines(existing_docs=[], overwrite=False, to_print=True)
        added_docs += self.get_docs_NewsArticles(existing_docs=[], overwrite=False, to_print=True)

        return added_docs

    def get_docs_Guidelines(self, existing_docs=[], overwrite=False, to_print=True):
        print("\n======================== Malta Guidelines =========================")
        added_docs = []

        page_url = 'https://idpc.org.mt/for-organisations/guidelines/'
        if to_print:
            print('Page:\t', page_url)
        page_source = self.get_source(page_url=page_url)
        if page_source is None:
            sys.exit("Couldn't connect to base url")

        results_soup = BeautifulSoup(page_source.text, 'html.parser')
        assert results_soup

        body = results_soup.find('body')
        assert body

        vce_text = body.find('div', class_='vce-text-block-wrapper', id='el-88463f26')
        assert vce_text

        iteration = 1
        for p_tag in vce_text.find_all('p'):
            assert p_tag

            # Check if the p_tag leads to link
            try:
                a_tag = p_tag.find('a')
                assert a_tag
            except:
                continue

            print('\n------------ Document: ' + str(iteration) + '-------------')
            iteration += 1

            # Concatenate p_tag and a_tag text as there is some inconsistency regarding which tag gets the title text
            document_title = p_tag.get_text()
            print('\tDocument Title: ' + document_title)
            document_hash = hashlib.md5(document_title.encode()).hexdigest()
            if document_hash in existing_docs and overwrite is False:
                if to_print:
                    print('\tSkipping existing document:\t', document_hash)
                continue

            document_href = a_tag.get('href')
            assert document_href

            # Check if the href contains the full link address
            if document_href.startswith('http') is False:
                # Remove the unnecessary first 5 characters from document_href
                href_stripped = document_href[4:]
                document_url = 'https://idpc.org.mt' + href_stripped
            else:
                document_url = document_href

            print('\tDocument URL: ' + document_url)
            if to_print:
                print("\tDocument: ", document_hash)

            document_response = None
            try:
                document_response = requests.request('GET', document_url)
                document_response.raise_for_status()
            except requests.exceptions.HTTPError as error:
                if to_print:
                    print(error)
                pass
            if document_response is None:
                continue

            dpa_folder = self.path
            document_folder = dpa_folder + '/' + 'Guidelines' + '/' + document_hash

            try:
                os.makedirs(document_folder)

                with open(document_folder + '/' + self.language_code + '.pdf', 'wb') as f:
                    f.write(document_response.content)

                with open(document_folder + '/' + self.language_code + '.txt', 'wb') as f:
                    try:
                        document_text = textract.process(document_folder + '/' + self.language_code + '.pdf')
                        f.write(document_text)
                    except:
                        pass

                with open(document_folder + '/' + 'metadata.json', 'w') as f:
                    metadata = {
                        'title': {
                            self.language_code: document_title
                        },
                        'md5': document_hash,
                        'releaseDate': 'Date not available',
                        'url': document_url
                    }
                    json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                added_docs.append(document_hash)
            except FileExistsError:
                print("\tDirectory path already exists, continue.")

        return added_docs

    def get_docs_NewsArticles(self, existing_docs=[], overwrite=False, to_print=True):
        print("\n======================== Malta News Articles =========================")
        added_docs = []

        # old link:
        # page_url = 'https://idpc.org.mt/news/'

        # new link:
        page_url = 'https://idpc.org.mt/news-latest/'
        if to_print:
            print('Page:\t', page_url)

        # Use selenium to get html parse object in order to show ALL html elements
        exec_path = WebdriverExecPolicy().get_system_path()
        options = webdriver.ChromeOptions()
        options.add_argument('headless')
        driver_doc = webdriver.Chrome(options=options, executable_path=exec_path)
        driver_doc.get(page_url)

        # The new link don't have "show more button", ignore below code
        '''
        # Click the show more button on the page
        show_button = driver_doc.find_element_by_class_name('page-btn')
        while show_button.text != 'No More posts':
            show_button.click()
            show_button = driver_doc.find_element_by_class_name('page-btn')

        time.sleep(5)
        '''

        # Now that we have view of all news articles, create parse object
        results_soup = BeautifulSoup(driver_doc.page_source, 'html.parser')
        assert results_soup

        body = results_soup.find('body')
        assert body

        container = body.find('div', class_='news-list-container')
        assert container

        iteration = 1
        for div in container.find_all('div', Recursive=False):
            assert div

            a_tag = div.find('a', class_='link')
            # Some div's don't have a_tag that contains href -> they are random webpage display elements
            try:
                assert a_tag
            except:
                continue

            print('\n------------ Document: ' + str(iteration) + '-------------')
            iteration += 1

            document_href = a_tag.get('href')
            assert document_href

            # Check if the href contains the full link address
            if document_href.startswith('http') is False:

                document_url = 'https://idpc.org.mt' + document_href
            else:
                document_url = document_href

            print('\tDocument URL: ' + document_url)

            document_response = None
            try:
                document_response = requests.request('GET', document_url)
                document_response.raise_for_status()
            except requests.exceptions.HTTPError as error:
                if to_print:
                    print(error)
                pass
            if document_response is None:
                continue

            document_soup = BeautifulSoup(document_response.text, 'html.parser')
            assert document_soup

            document_title = ''
            document_text = ''

            document_body = document_soup.find('body')
            assert document_body

            entry_content = document_body.find('div', class_='entry-content')
            assert entry_content

            # Want the second tag with class vce-row-container
            vce_row_list = entry_content.find_all('div', class_='vce-row-container')
            vce_row = vce_row_list[1]
            assert vce_row

            # This tag is what contains the entire document text
            vce_row_content = vce_row.find('div', class_='vce-row-content')
            assert vce_row_content

            # Now get the document title
            vce_row_two = vce_row_list[0]
            assert vce_row_two

            vce_content = vce_row_two.find('div', class_='vce-row-content')
            assert vce_content

            text_wrapper = vce_content.find('div', class_='vce-text-block-wrapper vce')
            assert text_wrapper

            # Set doc title and text
            document_title = text_wrapper.get_text()
            document_text = vce_row_content.get_text()

            print('\tDocument Title: ' + document_title)
            document_hash = hashlib.md5(document_title.encode()).hexdigest()
            if document_hash in existing_docs and overwrite is False:
                if to_print:
                    print('\tSkipping existing document:\t', document_hash)
                continue

            if to_print:
                print("\tDocument: ", document_hash)

            dpa_folder = self.path
            document_folder = dpa_folder + '/' + 'News Letters' + '/' + document_hash

            try:
                os.makedirs(document_folder)

                # Examine the document page for pdf links and download those encountered
                pdf_iteration = 1
                for p_tag in vce_row_content.find_all('p'):
                    assert p_tag
                    if p_tag.find('a') is not None:
                        a_tag = p_tag.find('a')
                        assert a_tag

                        href = a_tag.get('href')
                        assert href

                        # If we find a pdf link -> get the pdf
                        if href.endswith('.pdf'):
                            pdf_response = None
                            try:
                                pdf_response = requests.request('GET', href)
                                pdf_response.raise_for_status()
                            except requests.exceptions.HTTPError as error:
                                if to_print:
                                    print(error)
                                pass
                            if pdf_response is None:
                                continue

                            print("\tDownloading PDF: " + href)


                            with open(document_folder + '/' + self.language_code + str(pdf_iteration) + '.pdf', 'wb') as f:
                                f.write(pdf_response.content)
                            with open(document_folder + '/' + self.language_code + str(pdf_iteration) + '.txt', 'wb') as f:
                                link_text = textract.process(document_folder + '/' + self.language_code + str(pdf_iteration)
                                                             + '.pdf')
                                f.write(link_text)
                            pdf_iteration += 1
                    # No <a tags
                    else:
                        pass

                # Now download the document (summary) text
                with open(document_folder + '/' + self.language_code + 'Summary' + '.txt', 'wb') as f:
                    f.write(document_text.encode())

                with open(document_folder + '/' + 'metadata.json', 'w') as f:
                    metadata = {
                        'title': {
                            self.language_code: document_title
                        },
                        'md5': document_hash,
                        'releaseDate': 'Date not available',
                        'url': document_url
                    }
                    json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                added_docs.append(document_hash)
            except FileExistsError:
                print("\tDirectory path already exists, continue.")

        return added_docs



import os
import math
import requests
import json
import datetime
import hashlib
import dateparser
import re
from pygdpr.models.dpa import DPA
from bs4 import BeautifulSoup
from pygdpr.services.filename_from_path_service import filename_from_path_service
from pygdpr.services.pdf_to_text_service import PDFToTextService
from pygdpr.specifications import pdf_file_extension_specification
from pygdpr.specifications.should_retain_document_specification import ShouldRetainDocumentSpecification
from pygdpr.models.common.pagination import Pagination
from pygdpr.policies.gdpr_policy import GDPRPolicy
import textract
import time
import sys
import re

class Italy(DPA):
    def __init__(self, path=os.curdir):
        country_code='IT'
        super().__init__(country_code, path)

    def update_pagination(self, pagination=None, page_soup=None, driver=None, start_path=None):
        source = {
            "host": "https://www.garanteprivacy.it",
            "start_path": "/web/guest/home/stampa-comunicazione/interviste"
        }
        host = source['host']
        start_path = start_path

        if page_soup is not None:
            pagination = Pagination()

            # Page soup should be the page results_soup object
            pages = page_soup.find('ul', class_='pagination justify-content-center mt-3')
            assert pages

            li_page_list = pages.find_all('li', class_='page-item')
            assert li_page_list

            last_page_a = li_page_list[-2].find('a')
            assert last_page_a

            num_pages = int(last_page_a.get_text())

            # Add all the pages (including the first here)
            for num in range(1, num_pages + 1):
                page_link = host + start_path + str(num)
                # print(page_link)
                pagination.add_item(page_link)
        else:
            print("Please give update_pagination() a page_source argument")
            pass

        return pagination

    # Returns pagination list but the reverse of update_pagination
    def update_pagination_backwards(self, pagination=None, page_soup=None, driver=None, start_path=None):
        source = {
            "host": "https://www.garanteprivacy.it",
            "start_path": "/web/guest/home/stampa-comunicazione/interviste"
        }
        host = source['host']
        start_path = start_path

        if page_soup is not None:
            pagination = Pagination()

            # Page soup should be the page results_soup object
            pages = page_soup.find('ul', class_='pagination justify-content-center mt-3')
            assert pages

            li_page_list = pages.find_all('li', class_='page-item')
            assert li_page_list

            last_page_a = li_page_list[-2].find('a')
            assert last_page_a

            num_pages = int(last_page_a.get_text())

            # Add all the pages (including the first here)
            for num in range(num_pages, 0, -1):
                page_link = host + start_path + str(num)
                # print(page_link)
                pagination.add_item(page_link)
        else:
            print("Please give update_pagination() a page_source argument")
            pass

        return pagination

    def get_source(self, page_url=None, driver=None, to_print=True):
        assert (page_url is not None)
        results_response = None
        try:
            results_response = requests.request('GET', page_url, timeout=5)
            results_response.raise_for_status()
        except requests.exceptions.HTTPError as error:
            if to_print:
                print(error)
            pass
        return results_response

    def get_docs(self, existing_docs=[], overwrite=False, to_print=True):
        added_docs = []
        # call all the get_docs_X() functions
        added_docs += self.get_docs_Interviews(existing_docs=[], overwrite=False, to_print=True)
        added_docs += self.get_docs_Publications(existing_docs=[], overwrite=False, to_print=True)
        added_docs += self.get_docs_AnnualReports(existing_docs=[], overwrite=False, to_print=True)
        added_docs += self.get_docs_Newsletters(existing_docs=[], overwrite=False, to_print=True)
        added_docs += self.get_docs_Injunctions(existing_docs=[], overwrite=False, to_print=True)
        # added_docs += self.get_docs_InjunctionsFromLastPage(existing_docs=[], overwrite=False, to_print=True)
        added_docs += self.get_docs_Hearings(existing_docs=[], overwrite=False, to_print=True)

        return added_docs

    def get_docs_Interviews(self, existing_docs=[], overwrite=False, to_print=True):
        print("\n========================= Italy Decision Documents ===========================")
        added_docs = []

        page_url = 'https://www.garanteprivacy.it/home/stampa-comunicazione/interviste'
        if to_print:
            print('Page:\t', page_url)
        page_source = self.get_source(page_url=page_url)
        if page_source is None:
            sys.exit('page_source is None')
        results_soup = BeautifulSoup(page_source.text, 'html.parser')
        assert results_soup
        testo = results_soup.find('div', class_='testo')
        assert testo
        ul_all = testo.find_all('ul', recursive=False)

        # Use iterator to keep track of document number
        iterator = 1
        for ul in ul_all:
            for li in ul.find_all('li'):
                time.sleep(2)
                result_link = li.find('a')
                assert result_link

                print('\n------------ Document ' + str(iterator) + ' ------------')
                iterator += 1

                document_title = result_link.get_text()
                print('Document Title: ' + document_title)

                document_hash = hashlib.md5(document_title.encode()).hexdigest()
                print("\tdocument_hash: ", document_hash)
                if document_hash in existing_docs and overwrite is False:
                    if to_print:
                        print('\tSkipping existing document:\t', document_hash)
                    continue
                document_href = result_link.get('href')
                if document_href.startswith('http') and document_href.startswith("https://www.garanteprivacy.it") is False:
                    print('\tSkipping document that leads to page outside of dpa site')
                    continue
                document_url = document_href
                if document_href.startswith('http') is False:
                    host = "https://www.garanteprivacy.it"
                    document_url = host + document_url

                print('\tDocument URL: ' + document_url)

                document_response = None
                try:
                    document_response = requests.request('GET', document_url, timeout=10)
                    document_response.raise_for_status()
                except requests.exceptions.HTTPError as error:
                    if to_print:
                        print(error)
                    pass
                if document_response is None:
                    continue

                document_soup = BeautifulSoup(document_response.text, 'html.parser')
                assert document_soup

                # Get the document date:
                section = document_soup.find('section', id='content')
                assert section

                page_property = section.find('span', property='dc:date')

                # If a page is entirely blank, the first things the scraper notices is a failed assertion for property
                try:
                    assert page_property
                except AssertionError:
                    continue
                    print("Unable to obtain date from document page. It is likely blank")

                date = page_property.get_text()
                date_str = str(date)

                print('\tDocument Date: ' + date_str)

                # Weird bug where string length is much longer than it appear in html
                index = len(date_str)-11
                year_digits = date_str[index:]
                # print('\tDocument Year: ' + year_digits)
                date_str = date_str.strip()

                if int(year_digits) < 17:
                    print("Terminating program as only old document remain on page")
                    return added_docs

                if int(year_digits) < 18:
                    print("Skipping outdated document")
                    return added_docs

                dpa_folder = self.path
                document_folder = dpa_folder + '/' + 'Interviews' + '/' + document_hash
                # document_folder = dpa_folder + '/italy' + '/' + 'Interviews' + '/' + document_hash

                # Try to get document text directly from web page first
                try:
                    document_text = document_soup.find('div', id='div-to-print', class_='journal-content-article')
                    assert document_text

                    try:
                        os.makedirs(document_folder)

                        with open(document_folder + '/' + self.language_code + '.txt', 'wb') as f:
                            f.write(document_text.get_text().encode())

                        with open(document_folder + '/' + 'metadata.json', 'w') as f:
                            metadata = {
                                'title': {
                                    self.language_code: document_title
                                },
                                'md5': document_hash,
                                'releaseDate': date_str,
                                'url': document_url
                            }
                            json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                    except FileExistsError:
                        print("\tDirectory path already exists, continue.")

                # Try to download pdf instead
                except:
                    print("\tFailed to get text directly from page -> attempting to download pdf")
                    pdf_section = document_soup.find('section', id='content')
                    assert pdf_section

                    link_area = pdf_section.find('div', id='internal-content-wrapper')
                    assert link_area

                    pdf_a = link_area.find('a')
                    assert pdf_a

                    pdf_href = pdf_a.get('href')

                    if pdf_href.startswith('http') and pdf_href.startswith(host) is False:
                        continue
                    pdf_url = pdf_href
                    if pdf_href.startswith('http') is False:
                        host = "https://www.garanteprivacy.it"
                        pdf_url = host + pdf_url

                    print('\tPDF URL: ' + pdf_url)

                    if '.mp4' or '.mp3' in pdf_url:
                        print('\tLink leads to a video or audio file: Skipping')
                        continue

                    pdf_response = None
                    try:
                        pdf_response = requests.request('GET', pdf_url, timeout=5)
                        pdf_response.raise_for_status()
                    except requests.exceptions.HTTPError as error:
                        if to_print:
                            print(error)
                        pass
                    if pdf_response is None:
                        print("\tPDF response is None")
                        continue

                    if to_print:
                        print("\tDocument:\t", document_hash)

                    try:
                        os.makedirs(document_folder)

                        with open(document_folder + '/' + self.language_code + '.pdf', 'wb') as f:
                            f.write(pdf_response.content)
                        try:
                            with open(document_folder + '/' + self.language_code + '.txt', 'wb') as f:
                                pdf_text = textract.process(document_folder + '/' + self.language_code + '.pdf')
                                f.write(pdf_text)
                        except:
                            print("Failed to convert pdf to text document")

                        with open(document_folder + '/' + 'metadata.json', 'w') as f:
                            metadata = {
                                'title': {
                                    self.language_code: document_title
                                },
                                'md5': document_hash,
                                'releaseDate': date_str,
                                'url': document_url
                            }
                            json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                        added_docs.append(document_hash)
                    except FileExistsError:
                        print("\tDirectory path already exists, continue.")

        return added_docs
    

    def get_docs_Hearings(self, existing_docs=[], overwrite=False, to_print=True):
        print("\n========================= Italy Hearings ===========================")
        added_docs = []

        page_url = 'https://www.garanteprivacy.it/home/attivita-e-documenti/documenti/audizioni'
        if to_print:
            print('Page:\t', page_url)
        page_source = self.get_source(page_url=page_url)
        if page_source is None:
            sys.exit('page_source is None')

        results_soup = BeautifulSoup(page_source.text, 'html.parser')
        assert results_soup

        testo = results_soup.find('div', class_='testo')
        assert testo

        # Get p tags starting at third one in html layout
        p_all = testo.find_all('p', recursive=False)[3:]

        iteration = 1
        for p in p_all:

            print('\n------------- Document ' + str(iteration) + ' ------------')
            iteration +=1

            result_link = p.find('a')
            # Some p tag contain html for additional header separating the document links.
            # So if no <a tag is available, it's most likely one of these
            try:
                assert result_link
            except:
                continue

            title_date = p.find('em').get_text()
            assert title_date
            print(title_date)

            # Some of the titles are identical in the html since the date is separate by another tag.
            # So concatenate this date back on to get a unique document title.
            document_title = result_link.get_text() + title_date
            document_hash = hashlib.md5(document_title.encode()).hexdigest()
            if document_hash in existing_docs and overwrite is False:
                if to_print:
                    print('\tSkipping existing document:\t', document_hash)
                continue

            print('\tDocument Hash: ' + document_hash)

            document_href = result_link.get('href')
            if document_href.startswith('http') and document_href.startswith(host) is False:
                continue
            document_url = document_href

            if document_href.startswith('http') is False:
                host = "https://www.garanteprivacy.it"
                document_url = host + document_url

            print('\tdocument url: ' + document_url)

            document_response = None
            try:
                document_response = requests.request('GET', document_url, timeout=5)
                document_response.raise_for_status()
            except requests.exceptions.HTTPError as error:
                if to_print:
                    print(error)
                pass
            if document_response is None:
                continue

            document_soup = BeautifulSoup(document_response.text, 'html.parser')
            assert document_soup

            # Get the document date:
            section = document_soup.find('section', id='content')
            assert section

            property = section.find('span', property='dc:date')
            assert property

            date = property.get_text()
            date_str = str(date)

            print('\tDocument Date: ' + date_str)

            # Weird bug where string length is much longer than it appear in html
            index = len(date_str)-11
            year_digits = date_str[index:]
            #print('Year: ' + year_digits)

            if int(year_digits) < 17:
                print("Terminating program as only old document remain on page")
                return added_docs

            if int(year_digits) < 18:
                print("Skipping outdate document")
                return added_docs
            date_str = date_str.strip()
            dpa_folder = self.path
            document_folder = dpa_folder + '/' + 'Hearings' + '/' + document_hash
            # document_folder = dpa_folder + '/italy' + '/' + 'Hearings' + '/' + document_hash

            try:
                os.makedirs(document_folder)
                try:
                    document_text = document_soup.find('div', id='div-to-print')
                    assert document_text

                    # Should just find the very next div, which will get the scraper to the pure text
                    document_text_deeper = document_text.find('div')
                    assert document_text_deeper

                    with open(document_folder + '/' + self.language_code + '.txt', 'wb') as f:
                        f.write(document_text_deeper.get_text().encode())

                    with open(document_folder + '/' + 'metadata.json', 'w') as f:
                        metadata = {
                            'title': {
                                self.language_code: document_title
                            },
                            'md5': document_hash,
                            'releaseDate': date_str,
                            'url': document_url
                        }
                        json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                        print("\tDownloaded text")

                except:
                    pdf_section = document_soup.find('section', id='content')
                    assert pdf_section

                    link_area = pdf_section.find('div', id='internal-content-wrapper')
                    assert link_area

                    pdf_a = link_area.find('a')
                    assert pdf_a

                    pdf_href = pdf_a.get('href')

                    if pdf_href.startswith('http') and pdf_href.startswith(host) is False:
                        continue
                    pdf_url = pdf_href
                    if pdf_href.startswith('http') is False:
                        host = "https://www.garanteprivacy.it"
                        pdf_url = host + pdf_url

                    print('\tPDF URL: ' + pdf_url)

                    pdf_response = None
                    try:
                        pdf_response = requests.request('GET', pdf_url, timeout=5)
                        pdf_response.raise_for_status()
                    except requests.exceptions.HTTPError as error:
                        if to_print:
                            print(error)
                        pass
                    if pdf_response is None:
                        continue

                    if to_print:
                        print("\tDocument:", document_hash)

                    with open(document_folder + '/' + self.language_code + '.pdf', 'wb') as f:
                        f.write(pdf_response.content)
                    try:
                        with open(document_folder + '/' + self.language_code + '.txt', 'wb') as f:
                            pdf_text = textract.process(document_folder + '/' + self.language_code + '.pdf')
                            f.write(pdf_text)
                    except:
                        print("Failed to convert pdf to text document")

                    with open(document_folder + '/' + 'metadata.json', 'w') as f:
                        metadata = {
                            'title': {
                                self.language_code: document_title
                            },
                            'md5': document_hash,
                            'releaseDate': date_str,
                            'url': document_url
                        }
                        json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                        print("\tDownloaded pdf")
                added_docs.append(document_hash)
            except FileExistsError:
                print("\tDirectory path already exists, continue.")

        return added_docs

    # Utilizes the link starting at page 1, rather than the generic link
    def get_docs_Injunctions(self, existing_docs=[], overwrite=False, to_print=True):
        print("\n========================= Italy Injunctions ===========================")
        added_docs = []

        # Create pagination object and add all pages at once
        init_page_url = 'https://www.garanteprivacy.it/home/ricerca?p_p_id=g_gpdp5_search_GGpdp5SearchPortlet&p_p_lifecycle=0&p_p_state=normal&p_p_mode=view&_g_gpdp5_search_GGpdp5SearchPortlet_mvcRenderCommandName=%2FrenderSearch&_g_gpdp5_search_GGpdp5SearchPortlet_text=&_g_gpdp5_search_GGpdp5SearchPortlet_dataInizio=&_g_gpdp5_search_GGpdp5SearchPortlet_dataFine=&_g_gpdp5_search_GGpdp5SearchPortlet_idsTipologia=10526&_g_gpdp5_search_GGpdp5SearchPortlet_idsArgomenti=&_g_gpdp5_search_GGpdp5SearchPortlet_quanteParole=&_g_gpdp5_search_GGpdp5SearchPortlet_quanteParoleStr=&_g_gpdp5_search_GGpdp5SearchPortlet_nonParoleStr=&_g_gpdp5_search_GGpdp5SearchPortlet_paginaWeb=false&_g_gpdp5_search_GGpdp5SearchPortlet_allegato=false&_g_gpdp5_search_GGpdp5SearchPortlet_ordinamentoPer=DESC&_g_gpdp5_search_GGpdp5SearchPortlet_ordinamentoTipo=data&_g_gpdp5_search_GGpdp5SearchPortlet_cur=1'
        init_page_source = self.get_source(page_url=init_page_url)
        init_results_soup = BeautifulSoup(init_page_source.text, 'html.parser')
        assert init_results_soup
        # The start_path doesn't contain the very last character, which is the unique page number
        pagination = self.update_pagination(start_path='/home/ricerca?p_p_id=g_gpdp5_search_GGpdp5SearchPortlet&p_p_lifecycle=0&p_p_state=normal&p_p_mode=view&_g_gpdp5_search_GGpdp5SearchPortlet_mvcRenderCommandName=%2FrenderSearch&_g_gpdp5_search_GGpdp5SearchPortlet_text=&_g_gpdp5_search_GGpdp5SearchPortlet_dataInizio=&_g_gpdp5_search_GGpdp5SearchPortlet_dataFine=&_g_gpdp5_search_GGpdp5SearchPortlet_idsTipologia=10526&_g_gpdp5_search_GGpdp5SearchPortlet_idsArgomenti=&_g_gpdp5_search_GGpdp5SearchPortlet_quanteParole=&_g_gpdp5_search_GGpdp5SearchPortlet_quanteParoleStr=&_g_gpdp5_search_GGpdp5SearchPortlet_nonParoleStr=&_g_gpdp5_search_GGpdp5SearchPortlet_paginaWeb=false&_g_gpdp5_search_GGpdp5SearchPortlet_allegato=false&_g_gpdp5_search_GGpdp5SearchPortlet_ordinamentoPer=DESC&_g_gpdp5_search_GGpdp5SearchPortlet_ordinamentoTipo=data&_g_gpdp5_search_GGpdp5SearchPortlet_cur=', page_soup=init_results_soup)

        iteration = 1
        while pagination.has_next():
            page_url = pagination.get_next()
            if to_print:
                print('\n New Page:\t', page_url)

            page_source = self.get_source(page_url=page_url)
            if page_source is None:
                continue
            results_soup = BeautifulSoup(page_source.text, 'html.parser')
            assert results_soup

            section = results_soup.find('section', id='content')
            assert section

            blocco = section.find('div', class_='blocco-risultati')
            assert blocco

            for div in blocco.find_all('div', class_='card-risultato'):
                assert div
                time.sleep(2)

                # Obtain the document date
                date_div = div.find('div', class_='data-risultato')
                assert date_div
                p_tag = date_div.find('p')
                assert p_tag
                document_date = p_tag.get_text()

                # If we reach documents of year 2016 or below, stop scraping
                # We keep searching year 2017 just in case there are any remaining docs...
                document_year = document_date[-4:]
                if int(document_year) <= 2017:
                    print("Terminating because remaining documents are outdated")
                    # sys.exit("Remaining documents are outdated")
                    return added_docs

                # Check if the document is outdated
                if int(document_year) < 2018:
                    print("Skipping outdated document")
                    return added_docs

                result_link = div.find('a', class_='titolo-risultato')
                assert result_link

                print('\n------------ Document ' + str(iteration) + ' ------------')
                iteration += 1

                document_title = result_link.get_text()
                print('Document Title: ' + document_title)
                document_hash = hashlib.md5(document_title.encode()).hexdigest()
                print('\tDocument Date: ' + document_date)
                if document_hash in existing_docs and overwrite is False:
                    if to_print:
                        print('\tSkipping existing document:\t', document_hash)
                    continue

                document_href = result_link.get('href')
                if document_href.startswith('http') and document_href.startswith(host) is False:
                    continue
                document_url = document_href
                if document_href.startswith('http') is False:
                    host = "https://www.garanteprivacy.it"
                    document_url = host + document_url
                document_response = None

                try:
                    document_response = requests.request('GET', document_url, timeout=10)
                    document_response.raise_for_status()
                except requests.exceptions.HTTPError as error:
                    if to_print:
                        print(error)
                    pass
                if document_response is None:
                    continue
                print('\tdocument_url:\t', document_url)
                document_soup = BeautifulSoup(document_response.text, 'html.parser')
                assert document_soup

                body = document_soup.find('body')
                assert body

                body_text = body.find('div', class_='col-md-8 pl-4 px-md-5')
                assert body_text

                text_print_format = body_text.find('div', id='div-to-print')
                assert text_print_format

                second_text_print = text_print_format.find('div', class_='journal-content-article')

                if second_text_print is None:
                    second_text_print = text_print_format

                if to_print:
                    print("\tDocument:\t", document_hash)

                dpa_folder = self.path
                document_folder = dpa_folder + '/' + 'Injunctions' + '/' + document_hash
                # document_folder = dpa_folder + '/italy' + '/' + 'Injunctions' + '/' + document_hash

                try:
                    os.makedirs(document_folder)

                    with open(document_folder + '/' + self.language_code + '.txt', 'wb') as f:
                        f.write(second_text_print.get_text().encode())
                    with open(document_folder + '/' + 'metadata.json', 'w') as f:
                        metadata = {
                            'title': {
                                self.language_code: document_title
                            },
                            'md5': document_hash,
                            'releaseDate': document_date,
                            'url': document_url
                        }
                        json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                    added_docs.append(document_hash)
                except FileExistsError:
                    print("\tDirectory path already exists, continue.")
        return added_docs

    # This is an additional method to help scrape Injunctions Documents placed on the last pages of the website
    # Start at the last page of the injuctions link -> scrape backwards until reach outdated documents
    # This method keeps track of how many outdated documnets it encounters -> after hitting 15, stop scraping
    def get_docs_InjunctionsFromLastPage(self, existing_docs=[], overwrite=False, to_print=True):
        print("\n============ Italy Injunctions  (STARTING FROM LAST PAGE)===================")
        added_docs = []

        # Create pagination object and add all pages at once
        init_page_url = 'https://www.garanteprivacy.it/home/ricerca?p_p_id=g_gpdp5_search_GGpdp5SearchPortlet&p_p_lifecycle=0&p_p_state=normal&p_p_mode=view&_g_gpdp5_search_GGpdp5SearchPortlet_mvcRenderCommandName=%2FrenderSearch&_g_gpdp5_search_GGpdp5SearchPortlet_text=&_g_gpdp5_search_GGpdp5SearchPortlet_dataInizio=&_g_gpdp5_search_GGpdp5SearchPortlet_dataFine=&_g_gpdp5_search_GGpdp5SearchPortlet_idsTipologia=10526&_g_gpdp5_search_GGpdp5SearchPortlet_idsArgomenti=&_g_gpdp5_search_GGpdp5SearchPortlet_quanteParole=&_g_gpdp5_search_GGpdp5SearchPortlet_quanteParoleStr=&_g_gpdp5_search_GGpdp5SearchPortlet_nonParoleStr=&_g_gpdp5_search_GGpdp5SearchPortlet_paginaWeb=false&_g_gpdp5_search_GGpdp5SearchPortlet_allegato=false&_g_gpdp5_search_GGpdp5SearchPortlet_ordinamentoPer=DESC&_g_gpdp5_search_GGpdp5SearchPortlet_ordinamentoTipo=data&_g_gpdp5_search_GGpdp5SearchPortlet_cur=1'
        init_page_source = self.get_source(page_url=init_page_url)
        init_results_soup = BeautifulSoup(init_page_source.text, 'html.parser')
        assert init_results_soup
        # The start_path doesn't contain the very last character, which is the unique page number
        pagination = self.update_pagination_backwards(start_path='/home/ricerca?p_p_id=g_gpdp5_search_GGpdp5SearchPortlet&p_p_lifecycle=0&p_p_state=normal&p_p_mode=view&_g_gpdp5_search_GGpdp5SearchPortlet_mvcRenderCommandName=%2FrenderSearch&_g_gpdp5_search_GGpdp5SearchPortlet_text=&_g_gpdp5_search_GGpdp5SearchPortlet_dataInizio=&_g_gpdp5_search_GGpdp5SearchPortlet_dataFine=&_g_gpdp5_search_GGpdp5SearchPortlet_idsTipologia=10526&_g_gpdp5_search_GGpdp5SearchPortlet_idsArgomenti=&_g_gpdp5_search_GGpdp5SearchPortlet_quanteParole=&_g_gpdp5_search_GGpdp5SearchPortlet_quanteParoleStr=&_g_gpdp5_search_GGpdp5SearchPortlet_nonParoleStr=&_g_gpdp5_search_GGpdp5SearchPortlet_paginaWeb=false&_g_gpdp5_search_GGpdp5SearchPortlet_allegato=false&_g_gpdp5_search_GGpdp5SearchPortlet_ordinamentoPer=DESC&_g_gpdp5_search_GGpdp5SearchPortlet_ordinamentoTipo=data&_g_gpdp5_search_GGpdp5SearchPortlet_cur=', page_soup=init_results_soup)

        # outdate_count keeps track of how many outdated documents have been encountered so ar
        outdated_count = 0
        iteration = 1
        while pagination.has_next():
            page_url = pagination.get_next()
            if to_print:
                print('\n New Page:\t', page_url)
            page_source = self.get_source(page_url=page_url)
            if page_source is None:
                continue
            results_soup = BeautifulSoup(page_source.text, 'html.parser')
            assert results_soup

            section = results_soup.find('section', id='content')
            assert section

            blocco = section.find('div', class_='blocco-risultati')
            assert blocco

            for div in blocco.find_all('div', class_='card-risultato'):
                assert div
                time.sleep(2)

                # Obtain the document date
                date_div = div.find('div', class_='data-risultato')
                assert date_div
                p_tag = date_div.find('p')
                assert p_tag
                document_date = p_tag.get_text()

                result_link = div.find('a', class_='titolo-risultato')
                assert result_link

                # Try to get date from date tag, but if not, get the doc year from the title
                try:
                    # If failure occurs, will fail here
                    document_year = document_date[-4:]

                    # If we reach documents of year 2016 or below, stop scraping
                    # We keep searching year 2017 just in case there are any remaining docs...
                    if int(document_year) < 2018:
                        outdated_count += 1

                        if outdated_count >= 15:
                            print("Terminating because remaining documents are outdated")
                            # sys.exit("Remaining documents are outdated")
                            return added_docs
                        else:
                            print("Skipping outdated document")
                            continue

                # No date tag next to document title
                except:
                    title = result_link.get_text()
                    document_year = title[-14:-10]
                    document_date = document_year

                    # If we reach documents of year 2016 or below, stop scraping
                    # We keep searching year 2017 just in case there are any remaining docs...
                    if int(document_year) < 2018:
                        outdated_count += 1

                        if outdated_count >= 15:
                            print("Terminating because remaining documents are outdated")
                            # sys.exit("Remaining documents are outdated")
                            return added_docs

                        else:
                            print("Skipping outdated document")
                            continue

                print('\n------------ Document ' + str(iteration) + ' ------------')
                iteration += 1
                print('\tDocument Date: ' + document_date)
                document_title = result_link.get_text()
                print('\tDocument Title: ' + document_title)
                document_hash = hashlib.md5(document_title.encode()).hexdigest()
                if document_hash in existing_docs and overwrite is False:
                    if to_print:
                        print('\tSkipping existing document:\t', document_hash)
                    continue

                document_href = result_link.get('href')
                if document_href.startswith('http') and document_href.startswith(host) is False:
                    continue
                document_url = document_href
                if document_href.startswith('http') is False:
                    host = "https://www.garanteprivacy.it"
                    document_url = host + document_url
                document_response = None

                try:
                    document_response = requests.request('GET', document_url, timeout=5)
                    document_response.raise_for_status()
                except requests.exceptions.HTTPError as error:
                    if to_print:
                        print(error)
                    pass
                if document_response is None:
                    continue

                document_soup = BeautifulSoup(document_response.text, 'html.parser')
                assert document_soup

                body = document_soup.find('body')
                assert body

                body_text = body.find('div', class_='col-md-8 pl-4 px-md-5')
                assert body_text

                text_print_format = body_text.find('div', id='div-to-print')
                assert text_print_format

                second_text_print = text_print_format.find('div', class_='journal-content-article')
                assert second_text_print

                if to_print:
                    print("\tDocument:\t", document_hash)

                dpa_folder = self.path
                document_folder = dpa_folder + '/' + 'Injunctions Backwards' + '/' + document_hash
                # document_folder = dpa_folder + '/italy' + '/' + 'Injunctions Backwards' + '/' + document_hash

                try:
                    os.makedirs(document_folder)

                    with open(document_folder + '/' + self.language_code + '.txt', 'wb') as f:
                        f.write(second_text_print.get_text().encode())
                    with open(document_folder + '/' + 'metadata.json', 'w') as f:
                        metadata = {
                            'title': {
                                self.language_code: document_title
                            },
                            'md5': document_hash,
                            'releaseDate': document_date,
                            'url': document_url
                        }
                        json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                    added_docs.append(document_hash)
                except FileExistsError:
                    print("\tDirectory path already exists, continue.")
        return added_docs

    # Utilizes the link starting at page 1, rather than the generic link
    def get_docs_Newsletters(self, existing_docs=[], overwrite=False, to_print=True):
        print("\n========================== Italy Newsletters=============================")

        added_docs = []

        # Create pagination object and add all pages at once
        init_page_url = 'https://www.garanteprivacy.it/home/ricerca?p_p_id=g_gpdp5_search_GGpdp5SearchPortlet&p_p_lifecycle=0&p_p_state=normal&p_p_mode=view&_g_gpdp5_search_GGpdp5SearchPortlet_mvcRenderCommandName=%2FrenderSearch&_g_gpdp5_search_GGpdp5SearchPortlet_text=&_g_gpdp5_search_GGpdp5SearchPortlet_dataInizio=&_g_gpdp5_search_GGpdp5SearchPortlet_dataFine=&_g_gpdp5_search_GGpdp5SearchPortlet_idsTipologia=10524&_g_gpdp5_search_GGpdp5SearchPortlet_idsArgomenti=&_g_gpdp5_search_GGpdp5SearchPortlet_quanteParole=&_g_gpdp5_search_GGpdp5SearchPortlet_quanteParoleStr=&_g_gpdp5_search_GGpdp5SearchPortlet_nonParoleStr=&_g_gpdp5_search_GGpdp5SearchPortlet_paginaWeb=false&_g_gpdp5_search_GGpdp5SearchPortlet_allegato=false&_g_gpdp5_search_GGpdp5SearchPortlet_ordinamentoPer=DESC&_g_gpdp5_search_GGpdp5SearchPortlet_ordinamentoTipo=data&_g_gpdp5_search_GGpdp5SearchPortlet_cur=1'
        init_page_source = self.get_source(page_url=init_page_url)
        init_results_soup = BeautifulSoup(init_page_source.text, 'html.parser')
        assert init_results_soup
        # The start_path doesn't contain the very last character, which is the unique page number
        pagination = self.update_pagination(start_path='/home/ricerca?p_p_id=g_gpdp5_search_GGpdp5SearchPortlet&p_p_lifecycle=0&p_p_state=normal&p_p_mode=view&_g_gpdp5_search_GGpdp5SearchPortlet_mvcRenderCommandName=%2FrenderSearch&_g_gpdp5_search_GGpdp5SearchPortlet_text=&_g_gpdp5_search_GGpdp5SearchPortlet_dataInizio=&_g_gpdp5_search_GGpdp5SearchPortlet_dataFine=&_g_gpdp5_search_GGpdp5SearchPortlet_idsTipologia=10524&_g_gpdp5_search_GGpdp5SearchPortlet_idsArgomenti=&_g_gpdp5_search_GGpdp5SearchPortlet_quanteParole=&_g_gpdp5_search_GGpdp5SearchPortlet_quanteParoleStr=&_g_gpdp5_search_GGpdp5SearchPortlet_nonParoleStr=&_g_gpdp5_search_GGpdp5SearchPortlet_paginaWeb=false&_g_gpdp5_search_GGpdp5SearchPortlet_allegato=false&_g_gpdp5_search_GGpdp5SearchPortlet_ordinamentoPer=DESC&_g_gpdp5_search_GGpdp5SearchPortlet_ordinamentoTipo=data&_g_gpdp5_search_GGpdp5SearchPortlet_cur=', page_soup=init_results_soup)

        iteration = 1
        while pagination.has_next():
            page_url = pagination.get_next()
            if to_print:
                print('\n New Page:\t', page_url)
            page_source = self.get_source(page_url=page_url)
            if page_source is None:
                continue
            results_soup = BeautifulSoup(page_source.text, 'html.parser')
            assert results_soup

            section = results_soup.find('section', id='content')
            assert section

            blocco = section.find('div', class_='blocco-risultati')
            assert blocco

            for div in blocco.find_all('div', class_='card-risultato'):
                assert div
                time.sleep(2)

                # Obtain the document date
                date_div = div.find('div', class_='data-risultato')
                assert date_div
                p_tag = date_div.find('p')
                assert p_tag
                document_date = p_tag.get_text()

                # Check if the document is outdated or if we have reached a point where
                # won't encounter new documents
                document_year = document_date[-4:]

                if int(document_year) < 2017:
                    # sys.exit("Remaining documents are outdated")
                    # continue
                    return added_docs

                # Check if the document is outdated
                if int(document_year) < 2018:
                    print("Skipping outdated document")
                    return added_docs

                result_link = div.find('a', class_='titolo-risultato')
                assert result_link

                print('\n------------ Document ' + str(iteration) + ' ------------')
                iteration += 1
                print('\tDocument Date: ' + document_date)
                document_title = result_link.get_text()
                print('\tDocument Title: ' + document_title)
                document_hash = hashlib.md5(document_title.encode()).hexdigest()
                if document_hash in existing_docs and overwrite is False:
                    if to_print:
                        print('\tSkipping existing document:\t', document_hash)
                    continue

                document_href = result_link.get('href')
                if document_href.startswith('http') and document_href.startswith(host) is False:
                    continue
                document_url = document_href
                if document_href.startswith('http') is False:
                    host = "https://www.garanteprivacy.it"
                    document_url = host + document_url
                document_response = None

                try:
                    document_response = requests.request('GET', document_url, timeout=5)
                    document_response.raise_for_status()
                except requests.exceptions.HTTPError as error:
                    if to_print:
                        print(error)
                    pass
                if document_response is None:
                    continue

                document_soup = BeautifulSoup(document_response.text, 'html.parser')
                assert document_soup

                body = document_soup.find('body')
                assert body

                body_text = body.find('div', class_='col-md-8 pl-4 px-md-5')
                assert body_text

                text_print_format = body_text.find('div', id='div-to-print')
                assert text_print_format

                # Try to go deep in order to get text without excess html baggage
                obtained_text = None
                try:
                    second_text_print = text_print_format.find('div', class_='journal-content-article')
                    assert second_text_print
                    obtained_text = second_text_print
                except:
                    obtained_text = text_print_format

                assert obtained_text

                if to_print:
                    print("\tDocument:\t", document_hash)

                dpa_folder = self.path
                document_folder = dpa_folder + '/' + 'Newsletters' + '/' + document_hash
                # document_folder = dpa_folder + '/italy' + '/' + 'Newsletters' + '/' + document_hash

                try:
                    os.makedirs(document_folder)

                    with open(document_folder + '/' + self.language_code + '.txt', 'wb') as f:
                        f.write(obtained_text.get_text().encode())
                    with open(document_folder + '/' + 'metadata.json', 'w') as f:
                        metadata = {
                            'title': {
                                self.language_code: document_title
                            },
                            'md5': document_hash,
                            'releaseDate': document_date,
                            'url': document_url
                        }
                        json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                    added_docs.append(document_hash)
                except FileExistsError:
                    print("\tDirectory path already exists, continue.")
        return added_docs


    def get_docs_AnnualReports(self, existing_docs=[], overwrite=False, to_print=True):
        print("\n========================== Italy Annual Reports==========================")
        added_docs = []

        iteration = 1

        page_url = 'https://www.garanteprivacy.it/home/attivita-e-documenti/documenti/relazioni-annuali'

        if to_print:
            print('\n New Page:\t', page_url)
        page_source = self.get_source(page_url=page_url)
        if page_source is None:
            sys.exit("page_soure is None")
        results_soup = BeautifulSoup(page_source.text, 'html.parser')
        assert results_soup

        section = results_soup.find('section', id='content')
        assert section

        interna = section.find('div', class_='interna-webcontent')
        assert interna

        for tr in interna.find_all('tr')[1:]:
            assert tr

            # Obtain the document date
            td = tr.find_all('td')[0]
            assert td

            year = td.find('strong')

            assert year
            document_year = year.get_text()

            # Can use this to improve rutime since documents are already in chronological order
            if int(document_year) < 2017:
                break

            # Check if the document is outdated
            if int(document_year) < 2018:
                print("Skipping outdated document: " + document_year)
                continue

            result_link = tr.find('a')
            assert result_link

            print('\n------------ Document ' + str(iteration) + ' ------------')
            iteration += 1
            print('\tDocument Date: ' + document_year)
            document_title = "Annual Report " + document_year
            print('\tDocument Title: ' + document_title)
            document_hash = hashlib.md5(document_title.encode()).hexdigest()
            if document_hash in existing_docs and overwrite is False:
                if to_print:
                    print('\tSkipping existing document:\t', document_hash)
                continue

            document_href = result_link.get('href')
            if document_href.startswith('http') and document_href.startswith(host) is False:
                continue
            document_url = document_href
            if document_href.startswith('http') is False:
                host = "https://www.garanteprivacy.it"
                document_url = host + document_url
            document_response = None

            try:
                document_response = requests.request('GET', document_url, timeout=5)
                document_response.raise_for_status()
            except requests.exceptions.HTTPError as error:
                if to_print:
                    print(error)
                pass
            if document_response is None:
                continue

            document_soup = BeautifulSoup(document_response.text, 'html.parser')
            assert document_soup

            pdf_section = document_soup.find('section', id='content')
            assert pdf_section

            link_area = pdf_section.find('div', id='internal-content-wrapper')
            assert link_area

            pdf_a = link_area.find('a')
            assert pdf_a

            pdf_href = pdf_a.get('href')

            if pdf_href.startswith('http') and pdf_href.startswith(host) is False:
                continue
            pdf_url = pdf_href
            if pdf_href.startswith('http') is False:
                host = "https://www.garanteprivacy.it"
                pdf_url = host + pdf_url

            print('\tPDF URL: ' +pdf_url)

            pdf_response = None
            try:
                pdf_response = requests.request('GET', pdf_url, timeout=5)
                pdf_response.raise_for_status()
            except requests.exceptions.HTTPError as error:
                if to_print:
                    print(error)
                pass
            if pdf_response is None:
                continue

            if to_print:
                print("\tDocument:\t", document_hash)

            dpa_folder = self.path
            document_folder = dpa_folder + '/' + 'Annual Reports' + '/' + document_hash
            # document_folder = dpa_folder + '/italy' + '/' + 'Annual Reports' + '/' + document_hash

            try:
                os.makedirs(document_folder)

                with open(document_folder + '/' + self.language_code + '.pdf', 'wb') as f:
                    f.write(pdf_response.content)
                with open(document_folder + '/' + self.language_code + '.txt', 'wb') as f:
                    pdf_text = textract.process(document_folder + '/' + self.language_code + '.pdf')
                    f.write(pdf_text)
                with open(document_folder + '/' + 'metadata.json', 'w') as f:
                    metadata = {
                        'title': {
                            self.language_code: document_title
                        },
                        'md5': document_hash,
                        'releaseDate': document_year,
                        'url': document_url
                    }
                    json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                added_docs.append(document_hash)
            except FileExistsError:
                print("\tDirectory path already exists, continue.")
        return added_docs

    # Utilizes the link starting at page 1, rather than the generic link
    def get_docs_Publications(self, existing_docs=[], overwrite=False, to_print=True):
        print("\n========================== Italy Publications==========================")
        added_docs = []

        iteration = 1

        page_url = 'https://www.garanteprivacy.it/home/attivita-e-documenti/documenti/collana-contributi'
        if to_print:
            print('\n New Page:\t', page_url)
        page_source = self.get_source(page_url=page_url)
        if page_source is None:
            sys.exit('page_source is None')

        results_soup = BeautifulSoup(page_source.text, 'html.parser')
        assert results_soup

        section = results_soup.find('section', id='content')
        assert section

        testo = section.find('div', id='interna-webcontent')
        assert testo

        # Just get the very first <tr>
        tr = testo.find('tr')
        assert tr

        for p in tr.find_all('p'):
            assert p

            if p.find('br') is None:
                continue

            print('\n------------ Document ' + str(iteration) + ' ------------')
            iteration += 1

            document_title = p.get_text()
            print('\tDocument Title: ' + document_title)

            document_year = document_title[-4:]
            #print('\tDocument Year: ' + document_year)

            try:
                if int(document_year) < 2018:
                    print('Remaining documents are outdated')
                    break
            except ValueError:
                print("\tCouldn't parse document_year to integer")

            document_hash = hashlib.md5(document_title.encode()).hexdigest()
            if document_hash in existing_docs and overwrite is False:
                if to_print:
                    print('\tSkipping existing document:\t', document_hash)
                continue

            a_tag = p.find('a')
            assert a_tag

            href = a_tag.get('href')
            assert href

            document_url = 'https://www.garanteprivacy.it' + href
            print('\tDocument URL: ' + document_url)

            document_response = None
            try:
                document_response = requests.request('GET', document_url, timeout=5)
                document_response.raise_for_status()
            except requests.exceptions.HTTPError as error:
                if to_print:
                    print(error)
                pass
            if document_response is None:
                continue

            document_soup = BeautifulSoup(document_response.text, 'html.parser')
            assert document_soup

            document_section = document_soup.find('section', id='content')
            assert document_section

            portlet_body = document_section.find('div', class_='portlet-body')
            assert portlet_body

            content_wrapper = portlet_body.find('div', id='internal-content-wrapper')
            assert content_wrapper

            document_a = content_wrapper.find('a')
            assert document_a

            pdf_href = document_a.get('href')
            assert pdf_href

            pdf_url = 'https://www.garanteprivacy.it' + pdf_href
            print('\tPDF URL: ' + pdf_url)

            pdf_response = None
            try:
                pdf_response = requests.request('GET', pdf_url, timeout=5)
                pdf_response.raise_for_status()
            except requests.exceptions.HTTPError as error:
                if to_print:
                    print(error)
                pass
            if pdf_response is None:
                print('\tpdf_response is None')
                continue

            dpa_folder = self.path
            document_folder = dpa_folder + '/' + 'Publications' + '/' + document_hash
            # document_folder = dpa_folder + '/italy' + '/' + 'Publications' + '/' + document_hash

            try:
                os.makedirs(document_folder)

                with open(document_folder + '/' + self.language_code + '.pdf', 'wb') as f:
                    f.write(pdf_response.content)
                with open(document_folder + '/' + self.language_code + '.txt', 'wb') as f:
                    pdf_text = textract.process(document_folder + '/' + self.language_code + '.pdf')
                    f.write(pdf_text)
                with open(document_folder + '/' + 'metadata.json', 'w') as f:
                    metadata = {
                        'title': {
                            self.language_code: document_title
                        },
                        'md5': document_hash,
                        'releaseDate': document_year,
                        'url': pdf_url
                    }
                    json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                added_docs.append(document_hash)
            except FileExistsError:
                print("\tDirectory path already exists, continue.")
        return added_docs



from enum import Enum

class EUMember(Enum):
    """
    An Enum class used to map a country name (key) to it's country code (value).
    """
    AUSTRIA        = 'AT'
    BELGIUM        = 'BE'
    BULGARIA       = 'BG'
    CROATIA        = 'HR'
    CYPRUS         = 'CY'
    CZECH_REPUBLIC = 'CZ'
    DENMARK        = 'DK'
    ESTONIA        = 'EE'
    FINLAND        = 'FI'
    FRANCE         = 'FR'
    GERMANY        = 'DE'
    GREECE         = 'GR'
    HUNGARY        = 'HU'
    IRELAND        = 'IE'
    ITALY          = 'IT'
    LATVIA         = 'LV'
    LITHUANIA      = 'LT'
    LUXEMBOURG     = 'LU'
    MALTA          = 'MT'
    NETHERLANDS    = 'NL'
    POLAND         = 'PO'
    PORTUGAL       = 'PT'
    ROMANIA        = 'RO'
    SLOVAKIA       = 'SK'
    SLOVENIA       = 'SL'
    SPAIN          = 'ES'
    SWEDEN         = 'SE'
    UNITED_KINGDOM = 'GB'



import os
from pygdpr.services.eu_member_name_service import EUMemberNameService

class DPAPathPolicy:
    def extend_gdpr_path(self, gdpr_path, country_code):
        country_code = country_code.upper()
        eu_member_name = EUMemberNameService()
        try:
            country_name = eu_member_name.for_code(country_code)
        except:
            raise ValueError(f"Could not find matching country name for country_code: {country_code}")
        country_name = country_name.replace(' ', '-').lower()
        dpa_path = os.path.abspath(f"{gdpr_path}/{country_name}")
        return dpa_path



import os
import sys
import platform
from pygdpr.specifications.reachable_node_specification import ReachableNodeSpecification
from pygdpr.specifications.dpa_node_type_specification import DPANodeTypeSpecification
from pygdpr.services.github_issue_service import GithubIssueService
from pygdpr.services.markdown_formatting_service import MarkdownFormattingService
from pygdpr.services.py_line_inspector_service import PyLineInspectorService
from pygdpr.services.github_permalink_service import GithubPermalinkService
from pygdpr.services.github_commit_service import GithubCommitService

class ReportReachabilityIssueService:
    def get_title(self, dpa):
        country = dpa.country
        title = f"Unable to get docs from the DPA of {country}"
        return title

    def get_permalink(self, dpa):
        country = dpa.country
        github_commit = GithubCommitService()
        path = f"pygdpr/models/dpa/{country.lower()}/__init__.py"
        commit = github_commit.latest_commit(params={'path': path})
        if commit is None:
            return None
        py_line_inspector = PyLineInspectorService()
        lines = py_line_inspector.for_type_and_name(os.path.abspath(path), 'def', 'get_docs')
        github_permalink = GithubPermalinkService()
        permalink = github_permalink.generate(
            repo_owner=os.environ['gh-owner'],
            repo_name=os.environ['gh-name'],
            commit_id=commit['sha'],
            project_path=path,
            start_line=lines[0],
            end_line=lines[1]
        )
        return permalink

    def get_body(self, dpa):
        markdown_formatting = MarkdownFormattingService()
        print('permalink :: GITHUB')
        print(self.get_permalink(dpa))
        body = markdown_formatting.format_markdown(
            path=os.path.abspath('pygdpr/assets/gh-reachability-issue.md'),
            values=[
                ('host', 'https://wikileaks.org'),
                ('start_path', '/path'),
                ('country', dpa.country_code),
                ('major', sys.version_info.major),
                ('minor', sys.version_info.minor),
                ('micro', sys.version_info.micro),
                ('system', platform.system()),
                ('release', platform.release()),
                ('tests', 'Not written at the moment'),
                ('permalink', self.get_permalink(dpa))
            ],
            prefix='{{',
            suffix='}}'
        )
        return body

    def for_dpa_node_if_needed(self, dpa_node):
        dpa_node_type = DPANodeTypeSpecification()
        reachable_node = ReachableNodeSpecification()
        if dpa_node_type.is_satisfied_by(dpa_node) == False:
            raise ValueError("dpa_node argument is not of type 'DPANode'. Can only report a reachability issue for a dpa node")
        if reachable_node.is_satisfied_by(dpa_node):
            return None
        dpa = dpa_node.dpa
        github_issue = GithubIssueService()
        issue = None
        try:
            issue = github_issue.new_issue({
                "title": self.get_title(dpa),
                "body": self.get_body(dpa)
            })
        except:
            pass
        return issue



class GithubPermalinkService():
    def generate(self, repo_owner, repo_name, commit_id, project_path, start_line, end_line=None):
        lines = "L%d" % start_line
        if end_line is not None:
            lines += "-L%d" % end_line

        return f"https://github.com/{repo_owner}/{repo_name}/blob/{commit_id}/{project_path}#{lines}"



import os
import math
import requests
import json
import datetime
import hashlib
import dateparser
import re
from pygdpr.models.dpa import DPA
from bs4 import BeautifulSoup
from pygdpr.services.filename_from_path_service import filename_from_path_service
from pygdpr.services.pdf_to_text_service import PDFToTextService
from pygdpr.specifications import pdf_file_extension_specification
from pygdpr.specifications.should_retain_document_specification import ShouldRetainDocumentSpecification
from pygdpr.models.common.pagination import Pagination
from pygdpr.policies.gdpr_policy import GDPRPolicy
import textract
import time
from selenium import webdriver
from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from pygdpr.policies.webdriver_exec_policy import WebdriverExecPolicy
from selenium.common.exceptions import WebDriverException


class Ireland(DPA):
    # When running, make sure path is correct for you system
    def __init__(self, path=os.curdir):
        country_code='IE'
        super().__init__(country_code, path)

    def update_pagination(self, pagination=None, page_soup=None, driver=None):
        # ireland has two (official/primary) sources:
        # press releases and (latest) news.
        source = {
            "host": "https://www.dataprotection.ie",
            # "start_path": "/en/news-media/press-releases"
            "start_path": "/en/news-media/latest-news"
        }
        host = source['host']
        start_path = source['start_path']
        if pagination is None:
            pagination = Pagination()
            pagination.add_item(host + start_path)
        else:
            # pagination = Pagination()
            pager = page_soup.find('nav', class_='pager')
            if pager is None:
                return pagination
            pager_items = pager.find('ul', class_='pager__items')
            if pager_items is None:
                return pagination
            for pager_item in pager_items.find_all('li', 'pager__item'):
                page_link = pager_item.find('a')
                if page_link is None:
                    continue
                page_href = page_link.get('href')
                pagination.add_item(host + start_path + page_href)
                # print('added link to pagination:', host + start_path + page_href)
        return pagination

    # Fixed bug resulting in broken 'next page' links
    # Use this for get_docs_Blogs
    def update_pagination_noStart(self, pagination=None, page_soup=None, driver=None):
        # ireland has two (official/primary) sources:
        # press releases and (latest) news.
        source = {
            "host": "https://www.dataprotection.ie",
            # "start_path": "/en/news-media/press-releases"
            # "start_path": "/en/news-media/latest-news"  (This is the link the original scraper code uses)
            "start_path": "/dpc-guidance/blogs"
        }
        host = source['host']
        start_path = source['start_path']
        if pagination is None:
            pagination = Pagination()
            pagination.add_item(host + start_path)
        else:
            # pagination = Pagination()
            pager = page_soup.find('nav', class_='pager')
            if pager is None:
                return pagination
            pager_items = pager.find('ul', class_='pager__items')
            if pager_items is None:
                return pagination
            for pager_item in pager_items.find_all('li', 'pager__item'):
                page_link = pager_item.find('a')
                if page_link is None:
                    continue
                page_href = page_link.get('href')

                # Fixed Bug: Should be just host + page_href instead of host + start_path + page_href
                pagination.add_item(host + page_href)
                print('added link to pagination:', host + page_href)
        return pagination

    def get_source(self, page_url=None, driver=None):
        assert (page_url is not None)
        results_response = None
        try:
            results_response = requests.request('GET', page_url)
            results_response.raise_for_status()
        except requests.exceptions.HTTPError as error:
            if to_print:
                print(error)
            pass
        return results_response

    def get_docs(self, existing_docs=[], overwrite=False, to_print=True):
        added_docs = []
        # call all the get_docs_X() functions
        added_docs += self.get_docs_News(existing_docs=[], overwrite=False, to_print=True)
        added_docs += self.get_docs_Decisions(existing_docs=[], overwrite=False, to_print=True)
        added_docs += self.get_docs_Judgements(existing_docs=[], overwrite=False, to_print=True)
        added_docs += self.get_docs_Blogs(existing_docs=[], overwrite=False, to_print=True)
        added_docs += self.get_docs_Publications(existing_docs=[], overwrite=False, to_print=True)
        added_docs += self.get_docs_Guidances(existing_docs=[], overwrite=False, to_print=True)

        return added_docs


    def get_docs_News(self, existing_docs=[], overwrite=False, to_print=True):
        print("\n========================= Ireland New ===========================")
        iterator = 1

        existed_docs = []
        pagination = self.update_pagination()
        # s0. Pagination
        while pagination.has_next():
            page_url = pagination.get_next()
            if to_print:
                print('Page:\t', page_url)
            page_zero = 'https://www.dataprotection.ie/en/news-media/latest-news?page=0'
            if page_url == page_zero:
                print('page_url == page_zero')
                continue

            page_source = self.get_source(page_url=page_url)
            if page_source is None:
                continue
            results_soup = BeautifulSoup(page_source.text, 'html.parser')
            assert results_soup
            view_content = results_soup.find('div', class_='view-content')
            assert view_content
            item_list = view_content.find('div', class_='item-list')
            assert item_list
            ul = item_list.find('ul')
            assert ul
            # s1. Results
            for li in ul.find_all('li', recursive=False):
                print("\n------------ Document " + str(iterator) + " ------------")
                iterator += 1

                time.sleep(2)
                article = li.find('article')
                assert article
                p_date = article.find('p', class_='date')
                assert p_date
                date_str = p_date.get_text().strip()
                regex = r"(\d\d)(st|nd|rd|th) (\w*) (\d\d\d\d)"
                matches = re.finditer(regex, date_str)
                matches = list(matches)
                if len(matches) == 0:
                    continue
                match = matches[0]
                groups = match.groups()
                date_suffix_group_num = 2
                date_str = date_str[:match.start(date_suffix_group_num)] + date_str[match.end(date_suffix_group_num):]
                tmp = datetime.datetime.strptime(date_str, '%d %B %Y')
                date = datetime.date(tmp.year, tmp.month, tmp.day)
                if ShouldRetainDocumentSpecification().is_satisfied_by(date) is False:
                    #print('ShouldRetainDocumentSpecification is false')
                    continue
                h2 = article.find('h2')
                assert h2
                result_link = h2.find('a')
                assert result_link
                # s2. Documents
                document_title = result_link.get_text()
                print('\tDocument Title:\t', document_title)

                document_hash = hashlib.md5(document_title.encode()).hexdigest()
                if document_hash in existing_docs and overwrite == False:
                    if to_print:
                        print('\tSkipping existing document:\t', document_hash)
                    continue
                document_href = result_link.get('href')
                assert document_href
                host = "https://www.dataprotection.ie"
                document_url = host + document_href
                print('\tDocument Link:\t', document_url)

                if to_print:
                    print("\tDocument:\t", document_hash)
                document_response = None
                try:
                    document_response = requests.request('GET', document_url)
                    document_response.raise_for_status()
                except requests.exceptions.HTTPError as error:
                    if to_print:
                        print(error)
                    pass
                if document_response is None:
                    continue
                # Here is where we parse the document_url page (which is the new page containg the text or another link)
                # TODO: Write code to examine the pdf's, which are the publications
                document_soup = BeautifulSoup(document_response.text, 'html.parser')
                assert document_soup
                field_name_body = document_soup.find('div', class_='field--name-body')
                assert field_name_body

                document_text = field_name_body.get_text()
                dpa_folder = self.path
                #check whether have the same title
                if document_hash in existed_docs:

                    document_folder = dpa_folder + '/' + 'News' + '/' + document_hash + ' -02'
                else:

                    document_folder = dpa_folder + '/' + 'News' + '/' + document_hash
                try:
                    os.makedirs(document_folder)

                    with open(document_folder + '/' + self.language_code + '.txt', 'w') as f:
                        f.write(document_text)

                    with open(document_folder + '/' + 'metadata.json', 'w') as f:
                        metadata = {
                            'title': {
                                self.language_code: document_title
                            },
                            'md5': document_hash,
                            'releaseDate': date.strftime('%d/%m/%Y'),
                            'url': document_url
                        }
                        json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                    existed_docs.append(document_hash)

                except FileExistsError:
                    print("\tDirectory path already exists, continue.")

            pagination = self.update_pagination(pagination, page_soup=results_soup)

        return existed_docs

    def get_docs_Decisions(self, existing_docs=[], overwrite=False, to_print=True):
        print("\n========================= Ireland Decision Documents ===========================")
        iterator = 1

        existed_docs = []
        source = {
            "host": "https://www.dataprotection.ie",
            "start_path": "/en/dpc-guidance/law/decisions-made-under-data-protection-act-2018"
        }
        host = source['host']
        start_path = source['start_path']
        page_url = host + start_path
        page_source = self.get_source(page_url=page_url)
        print("decision: ",page_url)
        if page_source is None:
            print("This url is not exist.")

        results_soup = BeautifulSoup(page_source.text, 'html.parser')
        assert results_soup

        field_name_body = results_soup.find('div', class_='field--name-body')
        assert field_name_body

        for document in field_name_body.find_all('a'):
            assert document
            print("\n------------ Document " + str(iterator) + " ------------")
            iterator += 1

            time.sleep(2)
            document_title = document.get_text()
            print('\tDocument Title:\t', document_title)
            document_hash = hashlib.md5(document_title.encode()).hexdigest()
            print('\tdocument hash:\t', document_hash)
            date = document_title.split()[-2] + ' ' + document_title.split()[-1]
            default_day = '01st'

            full_date = default_day + ' ' + date
            date_str = full_date.strip()
            regex = r"(\d\d)(st|nd|rd|th) (\w*) (\d\d\d\d)"
            matches = re.finditer(regex, date_str)
            matches = list(matches)
            if len(matches) == 0:
                continue
            match = matches[0]
            groups = match.groups()
            date_suffix_group_num = 2
            date_str = date_str[:match.start(date_suffix_group_num)] + date_str[match.end(date_suffix_group_num):]
            tmp = datetime.datetime.strptime(date_str, '%d %B %Y')
            date = datetime.date(tmp.year, tmp.month, tmp.day)
            document_href = document.get('href')
            assert document_href
            document_link = host + document_href
            if document_href.startswith('https'):
                document_link = document_href
            print('\tdocument_link:\t', document_link)
            try:
                document_response = requests.request('GET', document_link)
                document_response.raise_for_status()
            except requests.exceptions.ConnectionError as error:
                pass

            if document_response is None:
                continue

            dpa_folder = self.path
            document_folder = dpa_folder + '/' + 'Decisions' + '/' + document_hash

            try:
                os.makedirs(document_folder)

                document_soup = BeautifulSoup(document_response.text, 'html.parser')
                field_name_body_2 = document_soup.find('div', class_='field--name-body')
                time.sleep(2)
                files = field_name_body_2.find_all('a')

                # get the inquiry file
                inquiry_href = files[0].get('href')
                inquiry_document_link = host + inquiry_href
                if inquiry_href.startswith('https'):
                    inquiry_document_link = inquiry_href
                inquiry_document_title = files[0].get_text()
                print('\t\tinquiry document link: ', inquiry_document_link)
                print('\t\tinquiry document title: ', inquiry_document_title)
                inquiry_file_response = None
                try:
                    inquiry_file_response = requests.request('GET', inquiry_document_link)
                    inquiry_file_response.raise_for_status()
                except requests.exceptions.ConnectionError as error:
                    pass
                if inquiry_file_response is None:
                    continue
                inquiry_file_content = inquiry_file_response.content

                # First, check inquiry type
                # the inquiry is .pdf type
                if inquiry_document_link.endswith('.pdf'):
                    with open(document_folder + '/' + self.language_code + '_Summary' + '.pdf', 'wb') as f:
                        f.write(inquiry_file_content)
                    with open(document_folder + '/' + self.language_code + '_Summary' + '.txt', 'w') as f:
                        inquiry_document_text = PDFToTextService().text_from_pdf_path(
                            document_folder + '/' + self.language_code + '_Summary' + '.pdf')
                        f.write(inquiry_document_text)

                # the inquiry is .txt type
                else:
                    content_soup = BeautifulSoup(inquiry_file_content, 'html.parser')
                    assert content_soup
                    inquiry_content = content_soup.find('div', class_='col-lg-9')
                    assert inquiry_content
                    short_document_text = inquiry_content.get_text()
                    with open(document_folder + '/' + self.language_code + '_Summary' + '.txt', 'w') as f:
                        f.write(inquiry_document_text)

                # Then, check number of files inside
                # contain multiple files
                if len(files) != 1:
                    count = 1
                    for file in files[1:]:
                        file_href = file.get('href')
                        file_link = host + file_href
                        if file_href.startswith('https'):
                            file_link = file_href
                        file_title = file.get_text()
                        print('\t\tfile link: ', file_link)
                        print('\t\tfile title: ',  file_title)
                        try:
                            file_response = requests.request('GET', file_link)
                            file_response.raise_for_status()
                        except requests.exceptions.ConnectionError as error:
                            pass
                        if file_response is None:
                            continue
                        file_content = file_response.content
                        with open(document_folder + '/' + self.language_code + '_Full-' + str(count) + '.pdf', 'wb') as f:
                            f.write(file_content)

                        # If a pdf fails to convert to text (because pdf is broken), print error
                        try:
                            with open(document_folder + '/' + self.language_code + '_Full-' + str(count) + '.txt', 'w') as f:
                                file_text = PDFToTextService().text_from_pdf_path(document_folder + '/' + self.language_code + '_Full' + '.pdf')
                                f.write(file_text)
                        except:
                            print("\t\tFailed to convert pdf to text document.")
                            pass
                        count += 1

                with open(document_folder + '/' + 'metadata.json', 'w') as f:
                    metadata = {
                        'title': {
                            self.language_code: document_title
                        },
                        'md5': document_hash,
                        'releaseDate': date.strftime('%m/%Y') + " --format: MM/YYYY",
                        'url': document_link
                    }
                    json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                existed_docs.append(document_hash)

            except FileExistsError:
                print("\tDirectory path already exists, continue.")

        return existed_docs

    def get_docs_Judgements(self, existing_docs=[], overwrite=False, to_print=True):
        print("\n========================= Ireland Judgements ===========================")

        # use selenium to bypass incapsula protection
        exec_path = WebdriverExecPolicy().get_system_path()
        options = webdriver.ChromeOptions()
        options.add_argument('headless')
        driver = webdriver.Chrome(options=options, executable_path=exec_path)

        existed_docs = []
        source = {
            "host": "https://www.dataprotection.ie",
            "start_path": "/en/dpc-guidance/law/judgments"
        }
        host = source['host']
        start_path = source['start_path']
        page_url = host + start_path
        print ("page :", page_url)

        assert (page_url is not None)
        results_response = None

        results_response = requests.request('GET', page_url)
        results_response

        results_soup = BeautifulSoup(results_response.text, 'html.parser')
        assert results_soup
        year_list = []
        for accordion in results_soup.find_all('div', class_='accordion'):
            for card in accordion.find_all('div', class_='card'):
                card_header = card.find('div', class_='card-header')

                year = card_header.find('h5', class_='mb-0').get_text()
                #print(year)

                if year < "2018" or year in year_list:
                    break
                year_list.append(year)
                card_body = card.find('div', class_='card-body')
                for articles in card_body.find_all('a'):

                    document_title = articles.get_text()
                    if document_title == '':
                        continue

                    # get the date of the document
                    date = document_title.split()[-3] + ' ' + document_title.split()[-2] + ' ' + document_title.split()[
                        -1]
                    # check whether contains specific day
                    if date[0] == '-':
                        date = '01 ' + date[2:]
                    # check whether contains year in title
                    if str(year) not in document_title:
                        # set a default date
                        date = '25 May 2018'
                    date_str = date.strip()

                    tmp = datetime.datetime.strptime(date_str, '%d %B %Y')
                    date = datetime.date(tmp.year, tmp.month, tmp.day)
                    if ShouldRetainDocumentSpecification().is_satisfied_by(date) is False:
                        #print('ShouldRetainDocumentSpecification is false')
                        continue

                    print('Document Title:\t', document_title)
                    document_href = articles.get('href')
                    document_hash = hashlib.md5(document_title.encode()).hexdigest()
                    dpa_folder = self.path
                    document_folder = dpa_folder + '/' + 'Judgements' + '/' + document_hash
                    try:
                        os.makedirs(document_folder)

                        if document_href.endswith('.html'):
                            #print("\t This is a html type")
                            document_url = document_href
                            print('\tDocument link:\t', document_url)

                            driver.get(document_url)
                            document_text = driver.find_element_by_xpath('//*[@id="search"]/div[2]/div/div[2]').text

                            with open(document_folder + '/' + self.language_code + '.txt', 'w') as f:
                                f.write(document_text)

                        elif document_href.endswith('/pdf'):
                            #print("\t This is a .pdf/pdf file")
                            document_url = document_href
                            document_response = None
                            try:
                                document_response = requests.request('GET', document_url)
                                document_response.raise_for_status()
                            except requests.exceptions.ConnectionError as error:
                                pass
                            if document_response is None:
                                continue
                            document_soup = BeautifulSoup(document_response.text, 'html.parser')
                            assert document_soup
                            full_view = document_soup.find('div', class_='download-text logo-pdf file-name')
                            full_view_href = full_view.find('a').get('href')
                            host_url = 'https://www.courts.ie/'
                            document_url = host_url + full_view_href
                            print('\tdocument_link: ', document_url)

                            full_view_response = None
                            try:
                                full_view_response = requests.request('GET', document_url)
                                full_view_response.raise_for_status()
                            except requests.exceptions.ConnectionError as error:
                                pass
                            if full_view_response is None:
                                continue
                            document_text = full_view_response.content
                            with open(document_folder + '/' + self.language_code + '.pdf', 'wb') as f:
                                f.write(document_text)
                            with open(document_folder + '/' + self.language_code + '.txt', 'w') as f:
                                document_text = PDFToTextService().text_from_pdf_path(
                                    document_folder + '/' + self.language_code + '.pdf')
                                f.write(document_text)

                        elif document_href.endswith('.pdf'):
                            print("\tThis is a PDF file")
                            document_url = host + document_href
                            print('\tdocument_link: ', document_url)
                            document_response = None
                            try:
                                document_response = requests.request('GET', document_url)
                                document_response.raise_for_status()
                            except requests.exceptions.ConnectionError as error:
                                pass
                            if document_response is None:
                                continue
                            document_text = document_response.content
                            with open(document_folder + '/' + self.language_code + '.pdf', 'wb') as f:
                                f.write(document_text)
                            with open(document_folder + '/' + self.language_code + '.txt', 'w') as f:
                                document_text = PDFToTextService().text_from_pdf_path(
                                    document_folder + '/' + self.language_code + '.pdf')
                                f.write(document_text)

                        else:
                            if document_href.startswith('https'):
                                document_url = document_href
                            else:
                                document_url = host + document_href
                            print('\tdocument_link: ', document_url)
                            document_response = None
                            try:
                                document_response = requests.request('GET', document_url)
                                document_response.raise_for_status()
                            except requests.exceptions.ConnectionError as error:
                                pass
                            if document_response is None:
                                continue
                            document_soup = BeautifulSoup(document_response.text, 'html.parser')
                            assert document_soup
                            field_name_body = document_soup.find('div', class_='field--name-body')
                            # assert field_name_body
                            if field_name_body is not None:
                                document_text = field_name_body.get_text()
                            else:
                                tab_Content = document_soup.find('div', class_='tabContent')
                                document_text = tab_Content.get_text()
                            with open(document_folder + '/' + self.language_code + '.txt', 'w') as f:
                                f.write(document_text)
                        print("\tDocument hash:\t", document_hash)
                        print('\tDocument URL:\t', document_url)

                        with open(document_folder + '/' + 'metadata.json', 'w') as f:
                            metadata = {
                                'title': {
                                    self.language_code: document_title
                                },
                                'md5': document_hash,
                                'releaseDate': date.strftime('%d/%m/%Y'),
                                'url': document_url
                            }
                            json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                        existed_docs.append(document_hash)
                        print('\n')
                    except FileExistsError:
                        print("\nDirectory path already exists, continue.")
        return existed_docs

    def get_docs_Blogs(self, existing_docs=[], overwrite=False, to_print=True):
        print("\n========================= Ireland Blogs ===========================")
        iterator = 1

        added_docs = []
        pagination = self.update_pagination_noStart()
        # s0. Pagination
        while pagination.has_next():
            page_url = pagination.get_next()
            if to_print:
                print('Page:\t', page_url)
            page_source = self.get_source(page_url=page_url)
            if page_source is None:
                continue
            results_soup = BeautifulSoup(page_source.text, 'html.parser')
            assert results_soup
            view_content = results_soup.find('div', class_='view-content')
            assert view_content

            # For https://www.dataprotection.ie/en/dpc-guidance/blogs, the articles are found under the
            # view_content class, each located at view-row. Thus we iterate through views-row to find data
            # for each article.

            # Iterate through view_content to get articles
            for li in view_content.find_all('div', recursive=False):
                time.sleep(2)
                article = li.find('article')
                assert article
                p_date = article.find('p', class_='date')
                assert p_date
                date_str = p_date.get_text().strip()
                regex = r"(\d\d)(st|nd|rd|th) (\w*) (\d\d\d\d)"

                matches = re.finditer(regex, date_str)
                matches = list(matches)
                if len(matches) == 0:
                    continue
                match = matches[0]
                groups = match.groups()
                date_suffix_group_num = 2
                date_str = date_str[:match.start(date_suffix_group_num)] + date_str[match.end(date_suffix_group_num):]
                tmp = datetime.datetime.strptime(date_str, '%d %B %Y')
                date = datetime.date(tmp.year, tmp.month, tmp.day)
                if ShouldRetainDocumentSpecification().is_satisfied_by(date) is False:
                    continue
                h2 = article.find('h2')
                assert h2
                result_link = h2.find('a')
                assert result_link
                # s2. Documents
                print('\n------------ Document: ' + str(iterator) + ' ------------')
                iterator += 1

                document_title = result_link.get_text()
                print('\tDocument Title\t', document_title)

                document_hash = hashlib.md5(document_title.encode()).hexdigest()
                if document_hash in existing_docs and overwrite == False:
                    if to_print:
                        print('\tSkipping existing document:\t', document_hash)
                    continue
                document_href = result_link.get('href')
                assert document_href
                host = "https://www.dataprotection.ie"
                document_url = host + document_href
                print('\tdocument_url:\t', document_url)
                if to_print:
                    print("\tDocument:\t", document_hash)
                document_response = None
                try:
                    document_response = requests.request('GET', document_url)
                    document_response.raise_for_status()
                except requests.exceptions.HTTPError as error:
                    if to_print:
                        print(error)
                    pass
                if document_response is None:
                    continue
                document_soup = BeautifulSoup(document_response.text, 'html.parser')
                assert document_soup
                field_name_body = document_soup.find('div', class_='field--name-body')
                assert field_name_body
                document_text = field_name_body.get_text()

                # Make sure that self.path is not the path of the most recent user who pushed to git
                dpa_folder = self.path

                # Set document folder as ireland/blogs/document_hash
                document_folder = dpa_folder + '/' + 'Blogs' + '/' + document_hash
                try:
                    os.makedirs(document_folder)

                    # Store files in ireland/blogs
                    with open(document_folder + '/' + self.language_code + '.txt', 'w') as f:
                        f.write(document_text)
                    with open(document_folder + '/' + 'metadata.json', 'w') as f:
                        metadata = {
                            'title': {
                                self.language_code: document_title
                            },
                            'md5': document_hash,
                            'releaseDate': date.strftime('%d/%m/%Y'),
                            'url': document_url
                        }
                        json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                    added_docs.append(document_hash)
                except FileExistsError:
                    print("\tDirectory path already exists, continue.")

            # Add link to the next page to pagination object
            pagination = self.update_pagination_noStart(pagination, page_soup=results_soup)

        return added_docs

    # Doesn't include financial accounts
    # This scraper method doesn't use update_pagination and update_pagination_noStart
    def get_docs_Publications(self, existing_docs=[], overwrite=False, to_print=True):
        print("\n========================= Ireland Publications ===========================")
        added_docs = []

        page_url = "https://www.dataprotection.ie/en/dpc-guidance/publications/annual-reports"
        if to_print:
            print('Page:\t', page_url)
        page_source = self.get_source(page_url=page_url)
        if page_source is None:
            sys.exit("Couldn't obtain page_source from page_url")
        results_soup = BeautifulSoup(page_source.text, 'html.parser')
        assert results_soup

        # view-content section contains articles within
        view_content = results_soup.find('div', class_='view-content')
        assert view_content

        # For each article on this page
        for article in view_content.find_all('article', recursive=True):
            time.sleep(2)
            article = article
            assert article
            p_date = article.find('p', class_='date')
            assert p_date
            date_str = p_date.get_text().strip()
            regex = r"(\d\d)(st|nd|rd|th) (\w*) (\d\d\d\d)"

            matches = re.finditer(regex, date_str)
            matches = list(matches)
            if len(matches) == 0:
                continue
            match = matches[0]
            groups = match.groups()
            date_suffix_group_num = 2
            date_str = date_str[:match.start(date_suffix_group_num)] + date_str[match.end(date_suffix_group_num):]
            tmp = datetime.datetime.strptime(date_str, '%d %B %Y')
            date = datetime.date(tmp.year, tmp.month, tmp.day)
            if ShouldRetainDocumentSpecification().is_satisfied_by(date) is False:
                continue
            h2 = article.find('h2')
            assert h2
            result_link = h2.find('a')
            assert result_link
            # s2. Documents
            document_title = result_link.get_text()
            print('Document Title:\t', document_title)

            document_hash = hashlib.md5(document_title.encode()).hexdigest()
            if document_hash in existing_docs and overwrite == False:
                if to_print:
                    print('\tSkipping existing document:\t', document_hash)
                continue

            document_href = result_link.get('href')
            assert document_href
            host = "https://www.dataprotection.ie"
            document_url = host + document_href
            if to_print:
                print("\tDocument Hash:\t", document_hash)
            document_response = None
            try:
                document_response = requests.request('GET', document_url)
                document_response.raise_for_status()
            except requests.exceptions.HTTPError as error:
                if to_print:
                    print(error)
                pass
            if document_response is None:
                continue

            # Create next page parse object -> visit pdf link from here
            document_soup = BeautifulSoup(document_response.text, 'html.parser')
            assert document_soup

            content = document_soup.find('div', class_='content')
            assert content
            pdf_article = content.find('div', class_='clearfix text-formatted field field--name-body field--type-text-with-summary field--label-hidden field__item')
            assert pdf_article
            pdf_link = pdf_article.find('a')
            assert pdf_link
            pdf_href = pdf_link.get('href')
            assert pdf_href

            # If the pdf_href contains the full link, then that is the pdf_url
            if "https://www.dataprotection.ie" in pdf_href:
                pdf_url = pdf_href
            # Otherwise, the href just has the extension to add on to the host link
            else:
                pdf_url = host + pdf_href
            #print(pdf_url)

            # Now try to open the pdf_url
            if to_print:
                print("\tPDF Document:\t", document_hash)
            pdf_response = None
            try:
                pdf_response = requests.request('GET', pdf_url)
                pdf_response.raise_for_status()
            except requests.exceptions.HTTPError as error:
                if to_print:
                    print(error)
                pass
            if pdf_response is None:
                continue

            # Make sure that self.path is not the path of the most recent user who pushed to git
            dpa_folder = self.path

            # For now, store the pdf in the folder for the page that links to the pdf
            document_folder = dpa_folder + '/' + 'Annual Reports' + '/' + document_hash

            try:
                os.makedirs(document_folder)

                # Write pdf_response contents to the pdf files in the folder
                with open(document_folder + '/' + self.language_code + '.pdf', 'wb') as f:
                    f.write(pdf_response.content)
                # Include a .txt file with pdf contents
                with open(document_folder + '/' + self.language_code + '.txt', 'w') as f:
                    full_document_text = PDFToTextService().text_from_pdf_path(
                        document_folder + '/' + self.language_code + '.pdf')
                    f.write(full_document_text)

                with open(document_folder + '/' + 'metadata.json', 'w') as f:
                    metadata = {
                        'title': {
                            self.language_code: document_title
                        },
                        'md5': document_hash,
                        'releaseDate': date.strftime('%d/%m/%Y'),
                        'url': document_url
                    }
                    json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                added_docs.append(document_hash)
            except FileExistsError:
                print("\tDirectory path already exists, continue.")
            print('\n')
        return added_docs

    # A big error message with this line: ConnectionRefusedError: [Errno 61] Connection refused,
    # probably means that there is an issue with href that causes the document url to be broken.

    # Doesn't visit any links that lead to external edpb.europa.eu or ec.europa.eu sites
    def get_docs_Guidances(self, existing_docs=[], overwrite=False, to_print=True):
        print("\n========================= Ireland Guidances ===========================")
        added_docs = []
        page_url = "https://www.dataprotection.ie/en/dpc-guidance"
        if to_print:
            print('Page:\t', page_url)
        page_source = self.get_source(page_url=page_url)
        if page_source is None:
            sys.exit("Couldn't obtain page_source from page_url")
        results_soup = BeautifulSoup(page_source.text, 'html.parser')
        assert results_soup
        view_content = results_soup.find('div', class_='card')
        assert view_content

        # Iterate through ul and li to get the articles (note, set recursive=True here to examine everything past scope)
        for ul in view_content.find_all('ul', recursive=True):
            # This will iterate through ALL li's -> even the ones within the scope of another li
            # (gets the very last links of the guidance page)
            for li in ul.find_all('li', recursive=True):
                time.sleep(2)

                # If there are multiple href links, get the second one only (first tends to be junk)
                if (len(li.find_all('a')) == 2):
                    article = li.find_all('a')[1]
                else:
                    article = li.find('a')

                assert article

                # The article object already contains the link
                result_link = article
                assert result_link
                # s2. Documents
                document_title = result_link.get_text()
                print('Document Title:\t', document_title)
                document_hash = hashlib.md5(document_title.encode()).hexdigest()
                if document_hash in existing_docs and overwrite == False:
                    if to_print:
                        print('\tSkipping existing document:\t', document_hash)
                    continue
                document_href = result_link.get('href')
                assert document_href

                # If 'href' is a link that goes to the external edpb or ec.europa site, skip it
                # Otherwise conditionally check if 'href' is a full link that starts with https://www.dataprotection.ie
                # If so, document_url = href, since the href is the full link and not just the extension
                if "https://edpb.europa.eu" in document_href:
                    continue
                if "ec.europa.eu" in document_href:
                    continue
                elif "https://www.dataprotection.ie" in document_href:
                    document_url = document_href
                else:
                    host = "https://www.dataprotection.ie"
                    document_url = host + document_href

                # Skip current iteration if forbidden url (right now, just have to match the exact url)
                # TODO: Provide way to check if link if forbidden in a non-hardcoded way
                if document_url == "https://www.dataprotection.ie/en/dpc-guidance/transfers-personal-data-ireland-uk-event-no-deal-brexit":
                    continue

                if to_print:
                    print("\tDocument Hash:\t", document_hash)
                document_response = None

                exec_path = WebdriverExecPolicy().get_system_path()
                options = webdriver.ChromeOptions()
                options.add_argument('headless')
                document_doc = webdriver.Chrome(options=options, executable_path=exec_path)
                try:
                    document_doc.get(document_url)
                    document_soup = BeautifulSoup(document_doc.page_source, 'html.parser')
                except WebDriverException:
                    print("connection refused")
                    pass

                # Look for pdf links
                # Set potential_link to empty be default
                potential_link = []
                try:
                    content = document_soup.find('div', class_='content')
                    assert content
                    potential_link_area = content.find('div', class_='clearfix text-formatted field field--name-body field--type-text-with-summary field--label-hidden field__item')
                    assert potential_link_area
                    potential_link = potential_link_area.find_all('a', recursive=True)
                    assert potential_link
                    for a in potential_link:
                        pdf_href_test = a.get('href')
                        assert pdf_href_test
                except AssertionError:
                    pass
                else:
                    potential_link = potential_link_area.find_all('a', recursive=True)

                # We have the pdf_response object -> now download its contents
                dpa_folder = self.path
                # For now, store the pdf in the folder for the page that links to the pdf
                document_folder = dpa_folder + '/' + "Guidances" + '/' + document_hash
                try:
                    os.makedirs(document_folder)

                    # If potential_link is not empty, iterate through its tag objects and look for pdf links
                    pdf_counter = 1
                    for a in potential_link:
                        pdf_href = a.get('href')
                        # Determine if the found link leads to a pdf
                        if isinstance(pdf_href, str):
                            if ".pdf" in pdf_href:
                                # If the pdf_href contains the full link, then that is the pdf_url
                                if "https://www.dataprotection.ie" in pdf_href:
                                    pdf_url = pdf_href
                                # Otherwise, the href just has the extension to add on to the host link
                                else:
                                    pdf_url = "https://www.dataprotection.ie" + pdf_href

                                # Broken links
                                if pdf_url == 'https://www.dataprotection.iehttps://uat.dataprotection.ie/sites/default/files/uploads/2020-12/Brexit%20FAQ_Dec20.pdf':
                                    continue
                                if pdf_url == "https://www.dataprotection.ie/sites/default/files/uploads/2020-04/Guide%20for%20Individuals%20Working%20Remotely_0_1.pdf":
                                    continue

                                # Now try to open the pdf_url
                                #if to_print:
                                    # TODO: Put pdf document hash here
                                    #print("\tPDF Document:\t", "place specific pdf doc hash here")

                                pdf_response = None
                                try:
                                    pdf_response = requests.request('GET', pdf_url)
                                    pdf_response.raise_for_status()
                                except requests.exceptions.HTTPError as error:
                                    if to_print:
                                        print(error)
                                    pass
                                # Except the other potential errors -> don't let it crash program
                                except:
                                    pass

                                if pdf_response is None:
                                    continue

                                    # Write pdf_response contents to the pdf files in the folder
                                    with open(document_folder + '/' + self.language_code + str(pdf_counter) + '.pdf', 'wb') as f:
                                        f.write(pdf_response.content)
                                    # Include a .txt file with pdf contents
                                    with open(document_folder + '/' + self.language_code + str(pdf_counter) + '.txt', 'w') as f:
                                        full_document_text = PDFToTextService().text_from_pdf_path(
                                            document_folder + '/' + self.language_code + str(pdf_counter) + '.pdf')
                                        f.write(full_document_text)

                                    pdf_counter = pdf_counter + 1
                                    added_docs.append(document_hash)

                            else:
                                pass
                        else:
                            pass

                    # Download text from the page as usual
                    field_name_body = document_soup.find('div', class_='field--name-body')
                    assert field_name_body
                    document_text = field_name_body.get_text()
                    with open(document_folder + '/' + self.language_code + 'Summary' + '.txt', 'w') as f:
                        f.write(document_text)
                    with open(document_folder + '/' + 'metadata.json', 'w') as f:
                        metadata = {
                            'title': {
                                self.language_code: document_title
                            },
                            'md5': document_hash,
                            # Can't use the "date" variable
                            'releaseDate': "Not available with current webpage",
                            'url': document_url
                        }
                        json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                    added_docs.append(document_hash)
                except FileExistsError:
                    print("\tDirectory path already exists, continue.")
                print('\n')
        return added_docs



import os
import json
import requests

class GithubIssueService:
    def get_auth(self):
        username, password = os.environ['gh-username'], os.environ['gh-password']
        return username, password

    def get_repo(self):
        owner, name = os.environ['gh-owner'], os.environ['gh-name']
        return owner, name

    def new_issue(self, content):
        session = requests.Session()
        username, password = self.get_auth()
        session.auth = (username, password)
        issue = {
            'title': content['title'],
            'body': content['body'] if 'body' in content else None,
            'milestone': content['milestone'] if 'milestone' in content else None,
            'labels': content['labels'] if 'labels' in content else [],
            'assignees': content['assignees'] if 'assignees' in content else []
        }
        print(json.dumps(issue))
        owner, name = self.get_repo()
        url = f"https://api.github.com/repos/{owner}/{name}/issues"
        response = session.post(url, json.dumps(issue))
        status_code = response.status_code
        print(status_code)
        if status_code != 201:
            raise ValueError('Could not create Github Issue with title: "%s"' % content['title'])
        return issue



import os
import json
import time
from pygdpr.specifications.supported_dpa_specification import SupportedDPASpecification
from pygdpr.specifications.eu_member_specification import EUMemberSpecification
from pygdpr.services.translate_price_service import TranslatePriceService
from pygdpr.services.translate_quota_service import TranslateQuotaService
from pygdpr.specifications.root_document_specification import RootDocumentSpecification
from pygdpr.policies.translate_file_policy import TranslateFilePolicy
from pygdpr.specifications.price_terminate_translate_specification import PriceTerminateTranslateSpecification
from pygdpr.specifications.not_reached_daily_translate_quota_specification import NotReachedDailyTranslateQuotaSpec
from pygdpr.specifications.not_reached_100_secs_translate_quota_specification import NotReached100SecsTranslateQuotaSpec
from pygdpr.specifications.translate_document_specification import TranslateDocumentSpecification
from pygdpr.specifications.not_exists_file_language_specification import NotExistsFileLanguageSpecification
#from google.cloud import translate_v2 as translate
import nltk
from nltk.corpus import stopwords
from nltk.stem import PorterStemmer
from sklearn.feature_extraction.text import CountVectorizer
from sklearn.feature_extraction.text import TfidfVectorizer
import pandas as pd
import matplotlib.pyplot as plt
from collections import Counter
from pygdpr.services.metadata.summary_metadata_service import *
from pygdpr.services.metadata.timeline_metadata_service import *
from pygdpr.services.metadata.citations_metadata_service import *
from pygdpr.services.metadata.citations_count_metadata_service import *
from pygdpr.services.metadata.keywords_metadata_service import *
from pygdpr.services.metadata.statistics_metadata_service import *
from pygdpr.services.metadata.est_read_time_metadata_service import *

nltk.download('punkt')
nltk.download('stopwords')
nltk.download('words')

supported_dpas = []
if os.path.isfile(os.path.abspath("pygdpr/assets/dpa-info.json")):
    with open(os.path.abspath("pygdpr/assets/dpa-info.json"), 'r') as f:
        supported_dpas = json.load(f)

class GoogleTranslatePriceError(Exception):
   """An exception class raised when the Google Translate price exceeds a predfined threshold (in usd)."""
   pass

class MaxRetriesError(Exception):
   """An exception class raised when maximum number of retries, trying to get docs, is exceeded."""
   pass

class DPA(object):
    """
    A class used to represent a DPA (Data Protection Authority).

    Attributes
    ----------
    country_code : str
        a two letter iso_code corresponding to a given country
    language_code : str
        a two letter iso_code referring to the written language 'preferred' by DPA
    name : str
        the official name of the DPA
    address : str
        formatted address string for the DPA
    phone : str
        formatted phone string for the DPA
    website : str
        official website of the DPA
    path : str
        the path where the documents, for this particular DPA, will be stored (default is os.cwd())
    translate_client : google.cloud.translate_v2.client.Client
        translate client for google-cloud-translate (see: https://googleapis.dev/python/translation/latest/client.html?highlight=client#module-google.cloud.translate_v2.client)

    Methods
    -------
    set_path(path)
        Sets the path to the given parameter. Stores documents at this path.
    set_translate_client(translate_client)
        Sets the translate_client to the given parameter. Client is used in :func:`~gdpr.dpa.DPA.translate_docs` to translate the DPA's documents at the specified path.
    get_docs(overwrite=False, to_print=True)
        Gets the documents for the DPA, stores them at the specified path and returns a list (md5 hashes) of documents added.
    translate_docs(target_languages, docs=[], overwrite=False, price_terminate_usd=0.0, quota_service=None, to_print=True)
        Translates the documents, located at the specified path, into the target languages and returns a list (md5 hashes) of documents translated.
    """
    def __init__(self, country_code, path):
        """
        Parameters
        ----------
        country_code : str
            The country_code for the country where the DPA has authority.
        """
        country_code = country_code.upper()
        if EUMemberSpecification().is_satisfied_by(country_code) is False:
            raise ValueError(f"Not found valid EU Member state for country code: {country_code}")
        if SupportedDPASpecification().is_satisfied_by(country_code) is False:
            raise ValueError(f"Not found supported DPA for country code: {country_code}")
        self.country_code = country_code
        dpa = supported_dpas[country_code]
        self.country = dpa['country']
        self.language_code = dpa['language_code']
        self.encoding = dpa['encoding'] if 'encoding' in dpa.keys() else None
        self.name = dpa['name']
        self.address = dpa['addressFormatted']
        self.phone = '({}) {}'.format(dpa['phone_code'], dpa['phone'])
        self.email = dpa['email']
        self.website = dpa['website']
        self.member = dpa['member']
        # Watch out for this path changing between users and the remote repo!!!
        self.path = path
        # self.path = '/Users/evanjacobs/Desktop/research_project/gdpr-sota/documents' + path
        # self.path = '/Users/chensun/PycharmProjects/gdpr-sota/documents' + path
        self.translate_client = None

    def set_path(self, path):
        self.path = path

    def set_translate_client(self, translate_client):
        self.translate_client = translate_client

    def update_pagination(self, pagination=None, page_soup=None, driver=None):
        """Updates and or creates a new pagination instance.

        Parameters
        ----------
        pagination : Pagination
            If pagination is None, a new pagination instance is created from scratch.
            Otherwise the pagination is updated with a new item.
        page_soup : Soup
            If page_soup is not None, this will be used to find and add the next item to the pagination.
        driver : Selenium
            If driver is not None, this will be used to find and add the next item to the pagination.

        Raises
        ------
        NotImplementedError
            If no subclass (ie. EU member state DPA) has implemented this method.

        Returns
        -------
        Pagination
            an updated pagination instance
        """
        raise NotImplementedError("'update_pagination' method not implemented.")

    def get_source(self, page_url=None, driver=None):
        """Returns a page source given either a page_url or a driver as input.

        Parameters
        ----------
        page_url : str
            If page_url is not None, this will be used to get the page source.
        driver : Selenium
            If driver is not None, this will be used to get the page source.

        Raises
        ------
        NotImplementedError
            If no subclass (ie. EU member state DPA) has implemented this method.

        Returns
        -------
        str
            a page source response from an html page.
        """
        raise NotImplementedError("'get_source' method not implemented.")

    def get_docs(self, existing_docs=[], overwrite=False, to_print=True):
        """Gets the documents for the DPA, stores them at the specified path and returns a list (md5 hashes) of documents added.

        Parameters
        ----------
        overwrite : bool
            If True, will overwrite the already existing documents at specified path.
            Otherwise will skip, if documents are already added.
        to_print : bool
            If True will print the progress of getting the documents.

        Raises
        ------
        NotImplementedError
            If no subclass (ie. EU member state DPA) has implemented this method.

        Returns
        -------
        [str]
            a list of documents (md5 hashes)
        """
        raise NotImplementedError("'get_docs' method not implemented.")

    def translate_docs(self, target_languages, docs=[], overwrite=False, price_terminate_usd=0.0, quota_service=None, to_print=True):
        """Translates the documents, located at the specified path, into the target languages and returns a list (md5 hashes) of documents translated.

        Parameters
        ----------
        target_languages : [str]
            A list of two letter iso_codes (language_codes) the documents will be translated into.
        docs : [str]
            A list of documents (md5 hashes) that needs to be translated. If list is empty it's assumed all documents at path should be translated.
        overwrite : bool
            If True, will overwrite the already existing documents at specified path.
            Otherwise will skip, if documents are already added.
        price_terminate_usd : float
            If price of translating the documents exceeds the price_terminate_usd parameter a GoogleTranslatePriceError will be raised.
            If price_terminate_usd is zero no GoogleTranslatePriceError will be raised.
        quota_service : TranslateQuotaService
            An instance of the TranslateQuotaService class which specifies the quota allowed for the google-cloud project with enabled google-cloud-translate api.
            If None is specified, default quota limits will be assumed. (See: https://cloud.google.com/translate/quotas)
        to_print : bool
            If True will print the progress of getting the documents.

        Returns
        -------
        [str]
            a list of translated documents (md5 hashes)
        """
        translated_docs = []
        target_languages = list(set(target_languages).difference(self.language_code))
        if self.translate_client is None:
            self.translate_client = translate.Client()
        price_service = TranslatePriceService()
        if quota_service is None:
            quota_service = TranslateQuotaService()
        agg_price = 0.0
        agg_quota = 0
        time_window_secs = 100
        for root,dirs,files in os.walk(self.path, topdown=True):
            if list(filter(lambda x: x.endswith('.txt'), files)) == 0:
                continue
            time.sleep(3)
            doc = root.split('/')[-1]
            if to_print and len(doc) > 0:
                print("Translating document:\t", doc)
                if doc in docs:
                    print("Progress:", docs.index(doc)/float(len(docs)))
            """if RootDocumentSpecification(root).is_satisfied_by(doc) is False:
                continue"""
            if TranslateDocumentSpecification(docs).is_satisfied_by(doc) is False:
                continue
            file_languages = list(filter(lambda x: len(x) == 2, [x.split('.')[0] for x in files]))
            for name in files:
                if TranslateFilePolicy().is_allowed(name) is False:
                    continue
                if PriceTerminateTranslateSpecification(price_terminate_usd).is_satisfied_by(agg_price):
                    raise GoogleTranslatePriceError(f"Estimated price of Google Translate exceeded value of {price_terminate_usd}.")
                with open(root + '/' + name, 'r', encoding='utf-8') as f:
                    document_text = f.read()
                    next_quota = agg_quota + len(document_text)
                    #if NotReachedDailyTranslateQuotaSpec(quota_service).is_satisfied_by(next_quota):
                    #    raise ValueError('Reached characters per day: %d. Please wait 24 hours until making another request.', agg_quota)
                    if NotReached100SecsTranslateQuotaSpec(quota_service).is_satisfied_by(next_quota):
                        if to_print:
                            print('Reached characters per 100 seconds per project per user: %d. sleeping %d secs before making next request.' % (agg_quota, time_window_secs))
                        time.sleep(time_window_secs+5)
                        agg_quota = 0
                    for target_language in target_languages:
                        if overwrite == True or\
                           NotExistsFileLanguageSpecification(file_languages).is_satisfied_by(target_language) == False:
                            continue
                        response = None
                        try:
                            # https://googleapis.dev/python/translation/latest/client.html
                            response = self.translate_client.translate(
                                document_text,
                                target_language=target_language,
                                format_='text' # text or html?
                            )
                        except:
                            if to_print:
                                print('Could not translate:\t', doc)
                            pass
                        if response is None:
                            continue
                        translated_text = response['translatedText']
                        with open(root + '/' + target_language + '.txt', 'w') as f:
                            f.write(translated_text)
                        agg_price += price_service.price_for_text(document_text)
                        agg_quota += len(document_text)
                        translated_docs.append(doc)
                        if to_print:
                            print('Translated:\t', doc)
                            print('\tFile:\t', name)
                            print('\tPrice:\t', agg_price)
        return translated_docs

    def extract_metadata(self, pipeline=[
        ('summary', SummaryMetadataService()),
        ('timeline', TimelineMetadataService()),
        ('citations', CitationsMetadataService()),
        ('citationsCount', CitationsCountMetadataService()),
        ('keywords', KeywordsMetadataService()),
        ('statistics', StatisticsMetadataService())
    ], docs=[], to_print=True):
        """Extracts the metadata from the documents.

        Parameters
        ----------
        pipeline : [tuple]
            A list of tuples where:
            First value in the tuple is a key of the metadata json file.
            Second value in the tuple is a value computed from the MetadataService method for_text
        docs : [str]
            A list of documents (md5 hashes) that requires metadata extraction. If list is empty it's assumed all documents at path should be translated.
        to_print : bool
            If True will print the progress of getting the documents.

        Returns
        -------
        [str]
            a list of documents (md5 hashes)
        """
        extracted_metadata = []
        for root, dirs, files in os.walk(self.path, topdown=True):
            print(root, dirs, files)
            split = root.split('/')
            if len(split) <= 4:
                continue
            country = split[-2]
            country = country.replace('-', ' ')
            country = ' '.join([word.capitalize() for word in country.split(' ')])
            if country != self.country:
                continue
            if 'en.txt' not in files:
                continue
            if 'metadata.json' not in files:
                continue
            if to_print:
                print('Country:\t', country)
            document_hash = split[-1]
            if document_hash not in docs:
                continue
            if to_print:
                print('Extracting metadata for:\t', document_hash)
            f = open(root + '/' + 'metadata.json', 'r')
            metadata = json.load(f)
            f.close()
            f = open(root + '/' + 'en.txt')
            document_text = f.read()
            f.close()
            for key, meta_service in pipeline:
                metadata[key] = meta_service.for_text(document_text)
            f = open(root + '/' + 'metadata.json', 'w')
            json.dump(metadata, f, indent=4, sort_keys=True)
            f.close()
            extracted_metadata.append(document_hash)
        return extracted_metadata



import os
import platform

class WebdriverExecPolicy():
    def get_system_path(self):
        system = platform.system()
        supported = ['Darwin', 'Linux', 'Windows']
        if system not in supported:
            raise ValueError('No chromedriver is supported for this platform system %s' % system)
        #return os.path.abspath("pygdpr/assets/chromedriver/%s" % system.lower())
        return os.path.abspath("pygdpr/assets/chromedriver")



class NotReached100SecsTranslateQuotaSpec():
    def __init__(self, quota_service):
        self.quota_service = quota_service
    def is_satisfied_by(self, quota_limit):
        return (quota_limit <= self.quota_service.get_chars_per_100_secs_per_user())



from pygdpr.models.common.reachability_node import Color

class ReachableNodeSpecification:
    def is_satisfied_by(self, node):
        reachable = True
        children = node.get_children()
        if len(children) == 0:
            return reachable
        colors = set([n.get_color() for n in children])
        if Color.WHITE in colors:
            reachable = False
        return reachable



import os
import math
import requests
import json
import time
import datetime
import hashlib
import dateparser
import re
import csv
from pygdpr.models.dpa import DPA
from bs4 import BeautifulSoup
from pygdpr.services.filename_from_path_service import filename_from_path_service
from pygdpr.services.pdf_to_text_service import PDFToTextService
from pygdpr.specifications import pdf_file_extension_specification
from pygdpr.specifications.should_retain_document_specification import ShouldRetainDocumentSpecification
from pygdpr.models.common.pagination import Pagination
from pygdpr.policies.gdpr_policy import GDPRPolicy
import textract
from selenium import webdriver
from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from pygdpr.policies.webdriver_exec_policy import WebdriverExecPolicy
import urllib3


class Sweden(DPA):
    def __init__(self, path=os.curdir):
        country_code='se'
        super().__init__(country_code, path)

    def update_pagination(self, pagination=None, page_soup=None, driver=None, start_path=None, current_page=1):
        source = {
            "host": "https://www.imy.se",
            "start_path_decisionsAndJudgements": "/tillsyner/",
            "start_path_publications": "/publikationer/"
        }
        host = source['host']
        if start_path == "decisionsAndJudgements":
            start_path = source['start_path_decisionsAndJudgements']
        else:
            start_path = source['start_path_publications']
        if pagination is None:
            pagination = Pagination()
            pagination.add_item(host + start_path + "?query=&page=1")
        else:
            pagination = Pagination()
            for i in range(current_page, 1000):
                page_href = "?query=&page={}".format(i)
                pagination.add_item(host + start_path + page_href)
        return pagination

    def get_source(self, page_url=None, driver=None):
        assert (page_url is not None)
        results_response = None
        try:
            results_response = requests.request('GET', page_url, verify=False)
            results_response.raise_for_status()
        except requests.exceptions.HTTPError as error:
            pass
        return results_response

    def get_docs(self, existing_docs=[], overwrite=False, to_print=True):
        added_docs = []
        # call all the get_docs_X() functions
        added_docs += self.get_docs_decisionsAndJudgements(existing_docs=[], overwrite=False, to_print=True)
        added_docs += self.get_docs_Publications(existing_docs=[], overwrite=False, to_print=True)
        added_docs += self.get_docs_Guidances(existing_docs=[], overwrite=False, to_print=True)
        return added_docs

    def get_docs_decisionsAndJudgements(self, existing_docs=[], overwrite=False, to_print=True):
        print("\n=============== Sweden Decision And Judgements =========================")
        urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
        iteration = 1
        existed_docs = []
        pagination = self.update_pagination(start_path="decisionsAndJudgements")
        # s0. Pagination
        while pagination.has_next():
            page_url = pagination.get_next()
            current_page = int(page_url[-1])
            if to_print:
                print('Page:\t', page_url)
            page_source = self.get_source(page_url=page_url)
            if page_source is None:
                return existed_docs
            results_soup = BeautifulSoup(page_source.text, 'html.parser')
            assert results_soup
            result_list = results_soup.find('ul', class_='imy-search__results-list')
            if result_list is None:
                print('End of the link')
                return existed_docs
            assert result_list
            # s1. Results
            for li in result_list.find_all('li', class_='imy-search__results-item'):
                time.sleep(2)
                item_header = li.find('h2', class_='imy-search-hit__heading')
                assert item_header

                # s2. Documents
                print('\n------------ Document ' + str(iteration) + ' ------------')
                iteration += 1
                document_title = item_header.get_text().strip()
                print('\tdocument_title:\t', document_title)
                document_hash = hashlib.md5(document_title.encode()).hexdigest()
                # get the date of the document
                hit_body = li.find('p', class_='imy-search-hit__body')
                assert hit_body
                document_summary = hit_body.get_text()
                if 'Beslut' not in document_summary:
                    print('\tThis is an ongoing decision')
                    continue
                date_str = document_summary.split()[-1:]
                date_str = ''.join(date_str)
                # date_str only contains year, set a default date
                if len(date_str) == 4:
                    date_str = date_str + '-01-01'
                tmp = dateparser.parse(date_str, date_formats=['%d %B %Y'])
                date = datetime.date(tmp.year, tmp.month, tmp.day)
                print('\tdate:\t', date)
                if ShouldRetainDocumentSpecification().is_satisfied_by(date) is False:
                    continue

                if document_hash in existing_docs and overwrite == False:
                    if to_print:
                        print('\tSkipping existing document:\t', document_hash)
                    continue
                item_link = li.find('a')
                assert item_link
                document_href = item_link.get('href')
                assert document_href
                document_url = document_href
                if to_print:
                    print("\tDocument:\t", document_hash)
                    print("\tdocument_url:\t", document_url)
                document_response = None
                try:
                    document_response = requests.request('GET', document_url, verify=False)
                    document_response.raise_for_status()
                except requests.exceptions.HTTPError as error:
                    if to_print:
                        print(error)
                    pass
                if document_response is None:
                    continue
                document_soup = BeautifulSoup(document_response.text, 'html.parser')
                assert document_soup
                body_content = document_soup.find('div', class_='imy-body imy-contentpage__main-content')
                assert body_content
                dpa_folder = self.path
                document_folder = dpa_folder + '/' + "Decisions & judgements" + '/' + document_hash
                try:
                    os.makedirs(document_folder)

                    file_href = None
                    file_links = [link for link in body_content.find_all('a') if link.get('href').endswith('.pdf')]
                    for i in range(len(file_links)):
                        file_number = i+1
                        file_href = file_links[i].get('href')
                        host = "https://www.imy.se"
                        file_url = host + file_href
                        print("\t\tfile_url:\t", file_url)
                        file_response = None
                        try:
                            file_response = requests.request('GET', file_url, verify=False)
                            file_response.raise_for_status()
                        except requests.exceptions.HTTPError as error:
                            pass
                        if file_response is None:
                            continue
                        file_content = file_response.content
                        try:
                            with open(document_folder + '/' + self.language_code + '_' + str(file_number) + '.pdf', 'wb') as f:
                                f.write(file_content)
                            with open(document_folder + '/' + self.language_code + '_' + str(file_number) + '.txt', 'wb') as f:
                                document_text = textract.process(document_folder + '/' + self.language_code + '_' + str(file_number) + '.pdf')
                                f.write(document_text)
                        except:
                            with open(document_folder + '/' + self.language_code + '.txt', 'w') as f:
                                document_text = newspage.get_text().strip()
                                f.write(document_text)
                    with open(document_folder + '/' + 'metadata.json', 'w') as f:
                        metadata = {
                            'title': {
                                self.language_code: document_title
                            },
                            'md5': document_hash,
                            'releaseDate': date.strftime('%d/%m/%Y'),
                            'url': document_url
                        }
                        json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                    existed_docs.append(document_hash)
                except FileExistsError:
                    print("\tDirectory path already exists, continue.")
            current_page += 1
            # s0. Pagination
            pagination = self.update_pagination(pagination=pagination, page_soup=results_soup, start_path='decisionsAndJudgements', current_page=current_page)
        return existed_docs

    def get_docs_Publications(self, existing_docs=[], overwrite=False, to_print=True):
        print("\n=============== Sweden Publications =========================")
        urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
        iteration = 1
        existed_docs = []
        pagination = self.update_pagination(start_path="publications")
        # s0. Pagination
        while pagination.has_next():
            page_url = pagination.get_next()
            current_page = int(page_url[-1])
            if to_print:
                print('Page:\t', page_url)
            page_source = self.get_source(page_url=page_url)
            if page_source is None:
                continue
            results_soup = BeautifulSoup(page_source.text, 'html.parser')
            assert results_soup
            result_list = results_soup.find('ul', {'id': 'imy-search__results-list-initial'})
            if result_list is None:
                print('End of the link')
                return existed_docs
            assert result_list
            # s1. Results
            for li in result_list.find_all('li', class_='imy-search__results-item'):
                time.sleep(2)
                item_created = li.find('time', class_='imy-search-hit__detail-text')
                assert item_created
                date_str = item_created.get('datetime')
                tmp = dateparser.parse(date_str, date_formats=['%d %B %Y'])
                date = datetime.date(tmp.year, tmp.month, tmp.day)
                if ShouldRetainDocumentSpecification().is_satisfied_by(date) is False:
                    continue
                item_header = li.find('h2', class_='imy-search-hit__heading')
                assert item_header
                # s2. Documents
                print('\n------------ Document ' + str(iteration) + ' ------------')
                iteration += 1
                document_title = item_header.get_text().strip()
                print('\tdocument_title:\t', document_title)
                document_hash = hashlib.md5(document_title.encode()).hexdigest()
                if document_hash in existing_docs and overwrite == False:
                    if to_print:
                        print('\tSkipping existing document:\t', document_hash)
                    continue
                item_link = li.find('a')
                assert item_link
                document_href = item_link.get('href')
                assert document_href
                document_url = document_href
                if to_print:
                    print("\tDocument:\t", document_hash)
                    print("\tdocument_url:\t", document_url)
                    print("\tdate:\t", date_str)
                document_response = None
                try:
                    document_response = requests.request('GET', document_url, verify=False)
                    document_response.raise_for_status()
                except requests.exceptions.HTTPError as error:
                    if to_print:
                        print(error)
                    pass
                if document_response is None:
                    continue
                document_soup = BeautifulSoup(document_response.text, 'html.parser')
                assert document_soup
                body_content = document_soup.find_all('div', class_='imy-info-block__content-container')
                assert body_content
                dpa_folder = self.path
                document_folder = dpa_folder + '/' + "Publications" + '/' + document_hash
                file_href = None
                file_link = body_content[-1].find('a', class_='imy-button')
                file_href = file_link.get('href')
                host = "https://www.imy.se"
                file_url = host + file_href
                # print('\tfile_url: ', file_url)
                file_response = None
                try:
                    file_response = requests.request('GET', file_url, verify=False)
                    file_response.raise_for_status()
                except requests.exceptions.HTTPError as error:
                    pass
                if file_response is None:
                    continue
                try:
                    os.makedirs(document_folder)

                    file_content = file_response.content
                    try:
                        with open(document_folder + '/' + self.language_code + '.pdf', 'wb') as f:
                            f.write(file_content)
                        with open(document_folder + '/' + self.language_code + '.txt', 'wb') as f:
                            document_text = textract.process(document_folder + '/' + self.language_code + '.pdf')
                            f.write(document_text)
                    except:
                        with open(document_folder + '/' + self.language_code + '.txt', 'w') as f:
                            document_text = newspage.get_text().strip()
                            f.write(document_text)
                    with open(document_folder + '/' + 'metadata.json', 'w') as f:
                        metadata = {
                            'title': {
                                self.language_code: document_title
                            },
                            'md5': document_hash,
                            'releaseDate': date.strftime('%d/%m/%Y'),
                            'url': document_url
                        }
                        json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                    existed_docs.append(document_hash)
                except FileExistsError:
                    print("\tDirectory path already exists, continue.")
            # s0. Pagination
            current_page += 1
            pagination = self.update_pagination(pagination=pagination, page_soup=results_soup, start_path="publications", current_page=current_page)
        return existed_docs

    def get_docs_Guidances(self, existing_docs=[], overwrite=False, to_print=True):
        print("\n=============== Sweden Guidances =========================")
        urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
        iteration = 1
        existed_docs = []
        source = {
            'host': 'https://www.imy.se',
            'start_path': '/verksamhet/dataskydd/dataskydd-pa-olika-omraden/',
        }
        host = source['host']
        start_path = source['start_path']
        page_url = host + start_path
        page_source = self.get_source(page_url=page_url)
        results_soup = BeautifulSoup(page_source.text, 'html.parser')
        assert results_soup
        exec_path = WebdriverExecPolicy().get_system_path()
        options = webdriver.ChromeOptions()
        options.add_argument('headless')
        driver_doc = webdriver.Chrome(options=options, executable_path=exec_path)
        driver_doc.get(page_url)
        guidance_section = driver_doc.find_element_by_xpath('//*[@id="readspeaker-content"]/div/div[2]/div[1]/nav/div/ul/li/ul/li[4]')
        for i in range(1, 9):
            guidance = guidance_section.find_element_by_xpath(
                '//*[@id="imy-content-menu__container-4a9aebd2-5a49-4b05-ab9c-ad1640c36a47"]/ul/li[' + str(i) + ']')
            title = guidance.find_element_by_xpath('//*[@id="imy-content-menu__container-4a9aebd2-5a49-4b05-ab9c-ad1640c36a47"]/ul/li['+ str(i)+']/div[1]/a/span')

            print('\n------------ Document ' + str(iteration) + ' ------------')
            iteration += 1

            document_title = title.get_attribute('textContent')
            document_hash = hashlib.md5(document_title.encode()).hexdigest()
            print("\tdocument title:\t", document_title)
            print("\tdocument hash:\t", document_hash)
            dpa_folder = self.path
            document_folder = dpa_folder + '/' + "Guidance" + '/' + document_hash
            try:
                os.makedirs(document_folder)

                elems = guidance.find_elements_by_tag_name('a')
                for j in range(len(elems)):
                    file_number = j+1
                    file_url = elems[j].get_attribute('href')
                    # the main url
                    exec_path = WebdriverExecPolicy().get_system_path()
                    options = webdriver.ChromeOptions()
                    options.add_argument('headless')
                    driver_file = webdriver.Chrome(options=options, executable_path=exec_path)
                    driver_file.get(file_url)
                    if j == 0:
                        document_url = file_url
                        print("\tdocument url:\t", document_url)
                        date_container = driver_file.find_element_by_class_name('imy-contentpage__date-container').get_attribute('textContent')
                        date_str = date_container.split(':')[-1].strip()
                        tmp = dateparser.parse(date_str, date_formats=['%d %B %Y'])
                        date = datetime.date(tmp.year, tmp.month, tmp.day)
                        print("\tdate:\t", date)
                        if ShouldRetainDocumentSpecification().is_satisfied_by(date) is False:
                            continue
                    try:
                        body_content = driver_file.find_element_by_xpath('//*[@id="readspeaker-content"]/div/div[2]/div[2]')
                    except:
                        body_content = driver_file.find_element_by_xpath('//*[@id="readspeaker-content"]/div[2]/div[2]/div/div[1]')
                    assert body_content
                    file_text = body_content.get_attribute('textContent')
                    with open(document_folder + '/' + self.language_code + '_' + str(file_number) + '.txt', 'w') as f:
                        f.write(file_text)
                with open(document_folder + '/' + 'metadata.json', 'w') as f:
                    metadata = {
                        'title': {
                            self.language_code: document_title
                        },
                        'md5': document_hash,
                        'releaseDate': date.strftime('%d/%m/%Y'),
                        'url': document_url
                    }
                    json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                existed_docs.append(document_hash)
            except FileExistsError:
                print("\tDirectory path already exists, continue.")
        return existed_docs





from pygdpr.services.metadata_service import MetadataService
import nltk

class EstReadTimeMetadataService(MetadataService):
    def for_text(self, text):
        # https://help.medium.com/hc/en-us/articles/214991667-Read-time
        wpm = 265
        words = nltk.word_tokenize(text)
        return len(words) / wpm



from enum import Enum

# Source: https://cloud.google.com/translate/pricing
class TranslateModelPrice(Enum):
    PBMT=20.0
    NMT=20.0
    AutoML=80.0

class TranslatePriceService():
    def price_for_text(self, text, model=TranslateModelPrice.PBMT):
        return model.value * (len(text) / 10.0**6)



import os
import errno
import json

class EUMemberSpecification():
    def is_satisfied_by(self, cand): # cand = country_code
        filename = 'eu-members.json'
        path = os.path.abspath(f"pygdpr/assets/{filename}")
        print("PATH: " + path)
        if os.path.isfile(path) is False:
            raise FileNotFoundError(
                errno.ENOENT, os.strerror(errno.ENOENT), filename
            )
        f = open(path, 'r')
        eu_members = json.load(f)
        f.close()
        country_codes = eu_members.keys()
        return cand.upper() in country_codes



class NotExistsFileLanguageSpecification():
    def __init__(self, file_languages):
        self.file_languages = file_languages
    def is_satisfied_by(self, file_language):
        return (file_language not in self.file_languages)



import os
from pygdpr.models.document.text import *
from pygdpr.models.document.metadata import *

class Document(object):
    def __init__(self, text, metadata):
        self.text = text
        self.metadata = metadata

    def read(self, path):
        return None

    def write(self, path):
        return None



import os
import shutil
import math
import time
import requests
import json
import hashlib
import datetime
from pygdpr.models.dpa import DPA
from bs4 import BeautifulSoup
from pygdpr.services.filename_from_path_service import filename_from_path_service
from pygdpr.services.pdf_to_text_service import PDFToTextService
from pygdpr.specifications import pdf_file_extension_specification
from pygdpr.models.common.pagination import Pagination
from pygdpr.specifications.should_retain_document_specification import ShouldRetainDocumentSpecification
import textract
from selenium import webdriver
from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from pygdpr.policies.webdriver_exec_policy import WebdriverExecPolicy
import urllib3

class Cyprus(DPA):
    def __init__(self, path=os.curdir):
        country_code='CY'
        super().__init__(country_code, path)

    def update_pagination(self, pagination=None, page_soup=None, start_path="Decisions"):
        if pagination is None:
            source = {
                "host": "http://www.dataprotection.gov.cy",
                "start_path_Decisions": "/DATAPROTECTION/DATAPROTECTION.NSF/dp06/dp06?opendocument",
                "start_path_AnnualReport": "/dataprotection/dataprotection.nsf/reports_gr/reports_gr?opendocument"
            }
            host = source['host']
            if start_path != "Decisions":
                start_path = source['start_path_AnnualReport']
            else:
                start_path = source['start_path_Decisions']
            pagination = Pagination()
            pagination.add_item(host + start_path)
        return pagination

    def get_source(self, page_url=None, driver=None):
        assert (page_url is not None)
        results_response = None
        try:
            results_response = requests.request('GET', page_url)
            results_response.raise_for_status()
        except requests.exceptions.HTTPError as error:
            if to_print:
                print(error)
            pass
        return results_response

    def get_docs(self, existing_docs=[], overwrite=False, to_print=True):
        added_docs = []
        # call all the get_docs_X() functions
        added_docs += self.get_docs_Decisions(existing_docs=[], overwrite=False, to_print=True)
        added_docs += self.get_docs_AnnualReports(existing_docs=[], overwrite=False, to_print=True)
        return added_docs

    def get_docs_Decisions(self, existing_docs=[], overwrite=False, to_print=True):
        existed_docs = []
        urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
        print("\n========================= Cyprus Decision Documents ===========================")
        # s0. Pagination
        pagination = self.update_pagination()
        while pagination.has_next():
            page_url = pagination.get_next()
            if to_print:
                print('Page:\t', page_url)
            page_source = self.get_source(page_url=page_url)
            if page_source is None:
                continue
            results_soup = BeautifulSoup(page_source.text, 'html.parser')
            assert results_soup
            content_block = results_soup.find('div', class_='content-block')
            assert content_block
            # s1. Results
            for li in content_block.find_all('li', class_='photos'):
                span_date = li.find('span', class_='date')
                assert span_date
                date_str = span_date.get_text()
                date_str = date_str.strip()
                tmp = datetime.datetime.strptime(date_str, '%d/%m/%Y')
                date = datetime.date(tmp.year, tmp.month, tmp.day)
                # print('document date: ', date)
                if ShouldRetainDocumentSpecification().is_satisfied_by(date) is False:
                    continue
                result_link = li.find('a')
                assert result_link
                # s2. Documents
                document_title = result_link.get_text()
                document_href = result_link.get('href')
                assert document_href
                host = "http://www.dataprotection.gov.cy"
                document_url = host + document_href
                document_response = None
                print(document_url)
                try:
                    document_response = requests.request('GET', document_url)
                    document_response.raise_for_status()
                except requests.exceptions.HTTPError as error:
                    if to_print:
                        print(error)
                    pass
                if document_response is None:
                    continue

                document_soup = BeautifulSoup(document_response.content, 'html.parser')
                assert document_soup

                # s3. detail pdf
                exec_path = WebdriverExecPolicy().get_system_path()
                options = webdriver.ChromeOptions()
                options.add_argument('headless')
                driver_doc = webdriver.Chrome(options=options, executable_path=exec_path)
                driver_doc.get(document_url)
                section = driver_doc.find_elements_by_class_name('simpletitle')

                for i in range(len(section)):
                    file_url = section[i].get_attribute('href')
                    if file_url.endswith('.pdf'):
                        time.sleep(1)
                        # print('\tfile_url: ', file_url)

                        file_title = section[i].text
                        if len(file_title) == 0:
                            file_title = document_title
                        print('file_title: ', file_title)
                        file_hash = hashlib.md5(file_title.encode()).hexdigest()
                        print('\tfile_hash: ', file_hash)

                        if file_hash in existing_docs and overwrite == False:
                            if to_print:
                                print('\tSkipping existing document:\t', file_hash)
                            continue

                        dpa_folder = self.path
                        file_folder = dpa_folder + '/' + 'Decisions' + '/' + file_hash
                        try:
                            os.makedirs(file_folder)
                            file_response = None
                            try:
                                file_response = requests.request('GET', file_url, verify=False)
                                file_response.raise_for_status()
                            except requests.exceptions.HTTPError as error:
                                if to_print:
                                    print(error)
                                pass
                            if file_response is None:
                                continue

                            file_content = file_response.content
                            if file_content is None:
                                continue
                            with open(file_folder + '/' + self.language_code + '.pdf', 'wb') as f:
                                f.write(file_content)
                            with open(file_folder + '/' + self.language_code + '.txt', 'wb') as f:
                                file_text = textract.process(file_folder + '/' + self.language_code + '.pdf')
                                f.write(file_text)
                            with open(file_folder + '/' + 'metadata.json', 'w') as f:
                                metadata = {
                                    'title': {
                                        self.language_code: file_title
                                    },
                                    'md5': file_hash,
                                    'releaseDate': 'Need to add',
                                    # 'releaseDate': date.strftime('%d/%m/%Y'),
                                    'url': file_url
                                }
                                json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                            existed_docs.append(file_hash)
                        except FileExistsError:
                            print("\tDirectory path already exists, continue.")

        return existed_docs

    def get_docs_AnnualReports(self, existing_docs=[], overwrite=False, to_print=True):
        existed_docs = []
        print("\n========================= Cyprus Annual Reports ===========================")
        # s0. Pagination
        pagination = self.update_pagination(start_path="AnnualReport")
        while pagination.has_next():
            page_url = pagination.get_next()
            if to_print:
                print('Page:\t', page_url)
            page_source = self.get_source(page_url=page_url)
            if page_source is None:
                continue
            results_soup = BeautifulSoup(page_source.text, 'html.parser')
            assert results_soup
            content_block = results_soup.find('div', class_='content-block')
            assert content_block
            # s1. Results
            for li in content_block.find_all('tr', {'valign': 'top'}):
                result_link = li.find('a')
                assert result_link
                # s2. Documents
                document_title = result_link.get_text()
                print("document_title: ", document_title)
                document_hash = hashlib.md5(document_title.encode()).hexdigest()
                if document_hash in existing_docs and overwrite == False:
                    if to_print:
                        print('\tSkipping existing document:\t', document_hash)
                    continue
                document_href = result_link.get('href')
                assert document_href
                host = "http://www.dataprotection.gov.cy"
                document_url = host + document_href
                # print('document_url: ', document_url)
                date = document_title.split()[-1]
                # print('\tdate: ', date)
                if date < '2018':
                    continue
                if to_print:
                    print("\tDocument:\t", document_hash)
                dpa_folder = self.path

                document_folder = dpa_folder + '/' + 'Annual Reports' + '/' + document_hash
                exec_path = WebdriverExecPolicy().get_system_path()
                options = webdriver.ChromeOptions()
                options.add_argument('headless')
                driver_doc = webdriver.Chrome(options=options, executable_path=exec_path)
                driver_doc.get(document_url)
                article_url = driver_doc.find_element_by_xpath('/html/body/form/section[3]/div/div/div/a').get_attribute("href")
                article_response = None
                try:
                    article_response = requests.request('GET', article_url)
                    article_response.raise_for_status()
                except requests.exceptions.HTTPError as error:
                    if to_print:
                        print(error)
                    pass
                if article_response is None:
                    continue
                document_content = article_response.content
                try:
                    os.makedirs(document_folder)

                    with open(document_folder + '/' + self.language_code + '.pdf', 'wb') as f:
                        f.write(document_content)
                    with open(document_folder + '/' + self.language_code + '.txt', 'w') as f:
                        document_text = PDFToTextService().text_from_pdf_path(
                            document_folder + '/' + self.language_code + '.pdf')
                        f.write(document_text)
                    with open(document_folder + '/' + 'metadata.json', 'w') as f:
                        metadata = {
                            'title': {
                                self.language_code: document_title
                            },
                            'md5': document_hash,
                            'releaseDate': date,
                            'url': document_url
                        }
                        json.dump(metadata, f, indent=4, sort_keys=True, ensure_ascii=False)
                    existed_docs.append(document_hash)
                except FileExistsError:
                    print("Directory path already exists, continue.")
        return existed_docs


import re
from pygdpr.services.metadata_service import *
from pygdpr.services.gdpr_chapter_section_service import *
import nltk
from nltk.stem.porter import *
from pygdpr.specifications.absolute_date_specification import *
from pygdpr.specifications.valid_gdpr_article_ref_specification import *
from dateparser.search import *

class CitationsMetadataService():
    def dot_notation_matches(self, words):
        refs = []
        # new regex to support cases: 14.a, 14.2
        regex = r"(\d\d*)\.((\d\d*)\.)?([a-z])"
        art_group_num, par_group_num, let_group_num = 1, 3, 4
        for w in words:
            matches = re.finditer(regex, w)
            for matchNum, match in enumerate(matches, start=1):
                art, par, let = None, None, None
                # print ("Match {matchNum} was found at {start}-{end}: {match}".format(matchNum = matchNum, start = match.start(), end = match.end(), match = match.group()))
                for groupNum in range(0, len(match.groups())):
                    groupNum = groupNum + 1
                    if match.group(groupNum) is None:
                        continue
                    if groupNum == art_group_num:
                        art = int(match.group(groupNum))
                    elif groupNum == par_group_num:
                        par = int(match.group(groupNum))
                    elif groupNum == let_group_num:
                        let = str(match.group(groupNum))
                    else:
                        continue
                    # print ("Group {groupNum} found at {start}-{end}: {group}".format(groupNum = groupNum, start = match.start(groupNum), end = match.end(groupNum), group = match.group(groupNum)))
                if art is None:
                    continue
                refs.append((art, par, let))
        return refs

    def for_text(self, text):
        ps = PorterStemmer()
        sents = nltk.sent_tokenize(text)
        art_key = ps.stem('article')
        par_key = ps.stem('paragraph')
        # letter key too, necessary?
        # let_key = ps.stem('letter')
        leg_key = '2016/679'
        for s in sents:
            ws = nltk.word_tokenize(s)
            ws = [w.lower() for w in ws]
            ws = [ps.stem(w) for w in ws]
            art_idx = -1
            leg_idx = -1
            try:
                art_idx = ws.index(art_key) # better to reuse: min(art_indices)?
            except:
                pass
            try:
                # leg_idx = ws.index(leg_key)
                leg_indices = [i for i, x in enumerate(ws) if x == leg_key]
                leg_idx = max(leg_indices)
            except:
                pass
            # if len(art_indices) == 0
            if art_idx == -1 or leg_idx == -1: # spec: gdprArticleRef.is_satisfied_by(s)
                continue
            #art_idx = art_indices[0]
            if art_idx > leg_idx: # support other case as well?
                continue
            ws_trim = ws[art_idx:leg_idx+1]
            matches = None
            try:
                matches = search_dates(" ".join(ws_trim), languages=['en'], settings={
                    'STRICT_PARSING': True,
                    'PREFER_DATES_FROM': 'past'
                })
            except:
                pass
            if matches is not None and len(matches) > 0:
                for m in matches:
                    if AbsoluteDateSpecification().is_satisfied_by(m):
                        snippet, date = m
                        snippet_ws = nltk.word_tokenize(snippet)
                        loc, len_ = -1, -1
                        for i in range(len(ws_trim)):
                            subset = ws_trim[i:i+len(snippet_ws)]
                            if subset == snippet_ws:
                                loc, len_ = i, i+len(snippet_ws)
                                break
                        if (loc != -1 and len_ != -1):
                            continue
                        del ws_trim[loc: len_]
            art_indices = [i for i, x in enumerate(ws_trim) if x == art_key]
            cand_arts = []
            for i in range(len(art_indices)):
                if i < len(art_indices)-1:
                    cand_arts.append(ws_trim[art_indices[i]:art_indices[i+1]])
                else:
                    cand_arts.append(ws_trim[art_indices[i]:])
            preds = []
            for cand in cand_arts:
                for i in range(1, len(cand)-1):
                    if cand[i].isdigit():
                        num = int(cand[i])
                        label = ''
                        if len(cand) >= 3 and (cand[i-1] == '(' and cand[i+1] == ')'):
                            label += 'par'
                        elif len(cand) >= 2 and (cand[i-1] == par_key):
                            label += 'par'
                        else:
                            label += 'art'
                        preds.append((num, label))
                    elif cand[i].isalpha() and len(cand[i]) == 1:
                        letter = str(cand[i])
                        label = 'let'
                        preds.append((letter, label))
                    else:
                        continue
            cand_refs = []
            for i in range(len(preds)):
                num, label = preds[i]
                if label == 'art':
                    cand_refs.append((num, None, None))
                elif label == 'par':
                    knear_art = -1
                    for j in range(len(preds[:i])):
                        if preds[j][1] == 'art':
                            knear_art = j
                    cand_art = preds[knear_art][0] if knear_art != -1 else None
                    cand_par = num
                    cand_let = None
                    cand_refs.append((cand_art, cand_par, cand_let))
                elif label == 'let':
                    knear_art, knear_par = -1, -1
                    for j in range(len(preds[:i])):
                        if preds[j][1] == 'art':
                            knear_art = j
                        elif preds[j][1] == 'par':
                            knear_par = j
                        else:
                            continue
                    cand_art = preds[knear_art][0] if knear_art != -1 else None
                    cand_par = preds[knear_par][0] if knear_par != -1 else None
                    cand_let = num
                    cand_refs.append((cand_art, cand_par, cand_let))
                else:
                    continue
            refs = []
            refs.extend(self.dot_notation_matches(ws_trim))
            i = 0
            while i < len(cand_refs):
                art, par, let = cand_refs[i]
                if i < len(cand_refs)-1:
                    if par is None and (cand_refs[i+1][0] == art and cand_refs[i+1][1] is not None):
                        refs.append(cand_refs[i+1])
                        i += 2
                    else:
                        refs.append(cand_refs[i])
                        i += 1
                else:
                    refs.append(cand_refs[i])
                    break
            final_refs = []
            j = 0
            while j < len(refs):
                art, par, let = refs[j]
                if j < len(refs)-1:
                    if let is None and ((refs[j+1][0] == art and refs[j+1][1] == par) and refs[j+1][2] is not None):
                        final_refs.append(refs[j+1])
                        j += 2
                    else:
                        final_refs.append(refs[j])
                        j += 1
                else:
                    final_refs.append(refs[j])
                    break
            valid_refs = []
            for ref in final_refs:
                if ValidGDPRArticleRefSpecification().is_satisfied_by(ref):
                    valid_refs.append(ref)
            article_refs = []
            for art, par, let in valid_refs:
                chapter, section = GDPRChapterSectionService().for_article(art)
                article_refs.append({
                    'chapter': chapter,
                    'section': section,
                    'article': art,
                    'paragraph': par,
                    'letter': let
                    # links['en'] = 'url' eur-lex.europa
                })
            return article_refs
